#!/usr/bin/env python
"""Tests for casing functions."""

from absl.testing import absltest

from grr_response_core.lib import casing


class SnakeToCamelTest(absltest.TestCase):

  def testSnakeToCamelWorksOnEmptyString(self):
    self.assertEqual("", casing.SnakeToCamel(""))

  def testSnakeToCamelWorksOnOneWordStrings(self):
    self.assertEqual("abcd", casing.SnakeToCamel("abcd"))
    self.assertEqual("a", casing.SnakeToCamel("a"))

  def testSnakeToCamelWorksOnRegularStrings(self):
    self.assertEqual("thisIsACamel", casing.SnakeToCamel("this_is_a_camel"))
    self.assertEqual("aCamelThisIs", casing.SnakeToCamel("a_camel_this_is"))
    self.assertEqual("aBCD", casing.SnakeToCamel("a_b_c_d"))

  def testSnakeToCamelWorksOnStringsWithUnderscoresOnly(self):
    self.assertEqual("", casing.SnakeToCamel("_"))
    self.assertEqual("", casing.SnakeToCamel("__"))
    self.assertEqual("", casing.SnakeToCamel("_________"))

  def testSnakeToCamelWorksOnStringsWithPrefixOrSuffixUnderscores(self):
    self.assertEqual("a", casing.SnakeToCamel("_a"))
    self.assertEqual("a", casing.SnakeToCamel("a_"))
    self.assertEqual("a", casing.SnakeToCamel("_a_"))
    self.assertEqual("a", casing.SnakeToCamel("___a___"))
    self.assertEqual("abcd", casing.SnakeToCamel("___abcd___"))
    self.assertEqual("aBCD", casing.SnakeToCamel("_a_b_c_d_"))
    self.assertEqual("aBCD", casing.SnakeToCamel("____a_b_c_d____"))
    self.assertEqual("aaBbCcDd", casing.SnakeToCamel("____aa_bb_cc_dd____"))
    self.assertEqual(
        "aaaBbbCccDdd", casing.SnakeToCamel("____aaa_bbb_ccc_ddd____")
    )

  def testSnakeToCamelWorksOnStringsWithMultipleUnderscoresBetweenWords(self):
    self.assertEqual("aBCD", casing.SnakeToCamel("a__b__c__d"))
    self.assertEqual("aBCD", casing.SnakeToCamel("a____b____c____d"))
    self.assertEqual("aBCD", casing.SnakeToCamel("___a___b___c___d___"))
    self.assertEqual("aaBbCcDd", casing.SnakeToCamel("___aa___bb___cc___dd___"))
    self.assertEqual(
        "aaaBbbCccDdd", casing.SnakeToCamel("___aaa___bbb___ccc___ddd___")
    )

  def testSnakeToCamelWorksOnStringsWithUppercaseLetters(self):
    self.assertEqual("a", casing.SnakeToCamel("A"))
    self.assertEqual("aaaaaa", casing.SnakeToCamel("Aaaaaa"))
    self.assertEqual("aaaaBbbbCccc", casing.SnakeToCamel("AaAa_bBbB_CcCc"))

  def testSnakeToCamelWorksOnStringsWithUnicodeCharacters(self):
    self.assertEqual("ąĆĘ", casing.SnakeToCamel("ą_ć_ę"))
    self.assertEqual("ąąąaĆććaĘęęa", casing.SnakeToCamel("ąąąa_ććća_ęęęa"))
    self.assertEqual("ąąąaĆććaĘęęa", casing.SnakeToCamel("ĄąĄA_ĆćĆa_ĘĘĘa"))


class CamelToSnakeTest(absltest.TestCase):

  def testCamelToSnakeWorksOnEmptyString(self):
    self.assertEqual("", casing.CamelToSnake(""))

  def testCamelToSnakeWorksOnOneWordStrings(self):
    self.assertEqual("abcd", casing.CamelToSnake("abcd"))
    self.assertEqual("a", casing.CamelToSnake("a"))

  def testCamelToSnakeWorksOnRegularStrings(self):
    self.assertEqual("this_is_a_snake", casing.CamelToSnake("thisIsASnake"))
    self.assertEqual("a_snake_this_is", casing.CamelToSnake("aSnakeThisIs"))
    self.assertEqual("a_b_c_d", casing.CamelToSnake("aBCD"))

  def testCamelToSnakeWorksOnStringsWithUppercaseLettersOnly(self):
    self.assertEqual(
        "t_h_i_s_i_s_a_s_n_a_k_e", casing.CamelToSnake("THISISASNAKE")
    )
    self.assertEqual(
        "a_s_n_a_k_e_t_h_i_s_i_s", casing.CamelToSnake("ASNAKETHISIS")
    )
    self.assertEqual("a_b_c_d", casing.CamelToSnake("ABCD"))
    self.assertEqual("a", casing.CamelToSnake("A"))

  def testCamelToSnakeWorksOnStringsWithUnicodeCharacters(self):
    self.assertEqual("ą_ć_ę", casing.CamelToSnake("ąĆĘ"))
    self.assertEqual("ąąąa_ććća_ęęęa", casing.CamelToSnake("ąąąaĆććaĘęęa"))
    self.assertEqual(
        "ą_ą_ąa_ć_ć_ća_ę_ę_ęa", casing.CamelToSnake("ĄĄĄaĆĆĆaĘĘĘa")
    )
    self.assertEqual("ą_ą_ą_ć_ć_ć_ę_ę_ę", casing.CamelToSnake("ĄĄĄĆĆĆĘĘĘ"))
    self.assertEqual("ą", casing.CamelToSnake("Ą"))

  def testCamelToSnakeWorksOnStringsWithUnderscores(self):
    self.assertEqual("a_b_c", casing.CamelToSnake("aB_c"))
    self.assertEqual("a_b", casing.CamelToSnake("a_b"))
    self.assertEqual("a_b", casing.CamelToSnake("A_b"))
    self.assertEqual("a_b", casing.CamelToSnake("a_B"))
    self.assertEqual("a_b", casing.CamelToSnake("A_B"))
    self.assertEqual("aa_bb", casing.CamelToSnake("_aaBb_"))
    self.assertEqual("a_bb_c", casing.CamelToSnake("___a_Bb__C___"))


if __name__ == "__main__":
  absltest.main()



#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from grr_response_core.lib.rdfvalues import paths as rdf_paths
from grr_response_proto import flows_pb2
from grr_response_proto import jobs_pb2


def ToProtoPathSpec(rdf: rdf_paths.PathSpec) -> jobs_pb2.PathSpec:
  return rdf.AsPrimitiveProto()


def ToRDFPathSpec(proto: jobs_pb2.PathSpec) -> rdf_paths.PathSpec:
  return rdf_paths.PathSpec.FromSerializedBytes(proto.SerializeToString())


def ToProtoGlobComponentExplanation(
    rdf: rdf_paths.GlobComponentExplanation,
) -> flows_pb2.GlobComponentExplanation:
  return rdf.AsPrimitiveProto()


def ToRDFGlobComponentExplanation(
    proto: flows_pb2.GlobComponentExplanation,
) -> rdf_paths.GlobComponentExplanation:
  return rdf_paths.GlobComponentExplanation.FromSerializedBytes(
      proto.SerializeToString()
  )



#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from grr_response_core.lib.rdfvalues import config_file as rdf_config_file
from grr_response_proto import config_file_pb2


def ToProtoLogTarget(
    rdf: rdf_config_file.LogTarget,
) -> config_file_pb2.LogTarget:
  return rdf.AsPrimitiveProto()


def ToRDFLogTarget(
    proto: config_file_pb2.LogTarget,
) -> rdf_config_file.LogTarget:
  return rdf_config_file.LogTarget.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoLogConfig(
    rdf: rdf_config_file.LogConfig,
) -> config_file_pb2.LogConfig:
  return rdf.AsPrimitiveProto()


def ToRDFLogConfig(
    proto: config_file_pb2.LogConfig,
) -> rdf_config_file.LogConfig:
  return rdf_config_file.LogConfig.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoNfsClient(
    rdf: rdf_config_file.NfsClient,
) -> config_file_pb2.NfsClient:
  return rdf.AsPrimitiveProto()


def ToRDFNfsClient(
    proto: config_file_pb2.NfsClient,
) -> rdf_config_file.NfsClient:
  return rdf_config_file.NfsClient.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoNfsExport(
    rdf: rdf_config_file.NfsExport,
) -> config_file_pb2.NfsExport:
  return rdf.AsPrimitiveProto()


def ToRDFNfsExport(
    proto: config_file_pb2.NfsExport,
) -> rdf_config_file.NfsExport:
  return rdf_config_file.NfsExport.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoNtpConfig(
    rdf: rdf_config_file.NtpConfig,
) -> config_file_pb2.NtpConfig:
  return rdf.AsPrimitiveProto()


def ToRDFNtpConfig(
    proto: config_file_pb2.NtpConfig,
) -> rdf_config_file.NtpConfig:
  return rdf_config_file.NtpConfig.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoPamConfigEntry(
    rdf: rdf_config_file.PamConfigEntry,
) -> config_file_pb2.PamConfigEntry:
  return rdf.AsPrimitiveProto()


def ToRDFPamConfigEntry(
    proto: config_file_pb2.PamConfigEntry,
) -> rdf_config_file.PamConfigEntry:
  return rdf_config_file.PamConfigEntry.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoPamConfig(
    rdf: rdf_config_file.PamConfig,
) -> config_file_pb2.PamConfig:
  return rdf.AsPrimitiveProto()


def ToRDFPamConfig(
    proto: config_file_pb2.PamConfig,
) -> rdf_config_file.PamConfig:
  return rdf_config_file.PamConfig.FromSerializedBytes(
      proto.SerializeToString()
  )



#!/usr/bin/env python
import collections

from absl.testing import absltest

from grr_response_core.lib.util import precondition


class AssertTypeTest(absltest.TestCase):

  def testIntCorrect(self):
    del self  # Unused.
    precondition.AssertType(108, int)
    precondition.AssertType(0xABC, int)
    precondition.AssertType(2**1024, int)

  def testIntIncorrect(self):
    with self.assertRaises(TypeError):
      precondition.AssertType(1.23, int)

    with self.assertRaises(TypeError):
      precondition.AssertType("123", int)

  def testStringCorrect(self):
    precondition.AssertType("foo", str)
    precondition.AssertType("gżegżółka", str)

  def testStringIncorrect(self):
    with self.assertRaises(TypeError):
      precondition.AssertType(b"foo", str)


class AssertOptionalTypeTest(absltest.TestCase):

  def testIntCorrect(self):
    del self  # Unused.
    precondition.AssertOptionalType(1337, int)
    precondition.AssertOptionalType(None, int)

  def testIntIncorrect(self):
    with self.assertRaises(TypeError):
      precondition.AssertOptionalType(13.37, int)

  def testStringCorrect(self):
    del self  # Unused.
    precondition.AssertOptionalType("foo", str)
    precondition.AssertOptionalType(None, str)

  def testStringIncorrect(self):
    with self.assertRaises(TypeError):
      precondition.AssertOptionalType(b"foo", str)

  def testBytesCorrect(self):
    del self  # Unused.
    precondition.AssertOptionalType(b"quux", bytes)
    precondition.AssertOptionalType(None, bytes)

  def testBytesIncorrect(self):
    with self.assertRaises(TypeError):
      precondition.AssertOptionalType("quux", bytes)


class AssertIterableTypeTest(absltest.TestCase):

  def testAssertEmptyCorrect(self):
    del self  # Unused.
    precondition.AssertIterableType([], int)
    precondition.AssertIterableType({}, str)

  def testStringSetCorrect(self):
    del self  # Unused.
    precondition.AssertIterableType({"foo", "bar", "baz"}, str)

  def testNonHomogeneousIntList(self):
    with self.assertRaises(TypeError):
      precondition.AssertIterableType([4, 8, 15, 16.0, 23, 42], int)

  def testIteratorIsNotIterable(self):
    with self.assertRaises(TypeError):
      precondition.AssertIterableType(iter(["foo", "bar", "baz"]), str)

  def testGeneratorIsNotIterable(self):

    def Generator():
      yield 1
      yield 2
      yield 3

    with self.assertRaises(TypeError):
      precondition.AssertIterableType(Generator(), int)


class AssertDictTypeTest(absltest.TestCase):

  def testIntStringDictCorrect(self):
    del self  # Unused.
    dct = {1: "foo", 2: "bar", 3: "baz"}
    precondition.AssertDictType(dct, int, str)

  def testNotADictIncorrect(self):
    dct = [(1, "foo"), (2, "bar"), (3, "baz")]
    with self.assertRaises(TypeError):
      precondition.AssertDictType(dct, int, str)

  def testWrongKeyType(self):
    dct = {"foo": 1, b"bar": 2, "baz": 3}
    with self.assertRaises(TypeError):
      precondition.AssertDictType(dct, str, int)

  def testWrongValueType(self):
    dct = {"foo": 1, "bar": 2, "baz": 3.14}
    with self.assertRaises(TypeError):
      precondition.AssertDictType(dct, str, int)

  def testSubclassesOfDictPassAssertion(self):
    dct = collections.UserDict({1: "foo", 2: "bar", 3: "baz"})
    precondition.AssertDictType(dct, int, str)

  def testSubclassesOfKeyAndValuePassAssertion(self):

    class A:
      pass

    class B(A):
      pass

    dct = {B(): B()}
    precondition.AssertDictType(dct, A, A)


if __name__ == "__main__":
  absltest.main()



#!/usr/bin/env python
"""Tests for the metrics interface for stats collection."""

from unittest import mock

from absl.testing import absltest

from grr_response_core.stats import default_stats_collector
from grr_response_core.stats import metrics
from grr_response_core.stats import stats_collector_instance
from grr.test_lib import stats_test_lib


class MetricsTest(
    stats_test_lib.StatsTestMixin,
    stats_test_lib.StatsCollectorTestMixin,
    absltest.TestCase,
):

  def testCounterRegistration(self):
    with self.SetUpStatsCollector(
        default_stats_collector.DefaultStatsCollector()
    ):
      metrics.Counter("cfoo")
    self.assertIsNotNone(self.collector.GetMetricMetadata("cfoo"))

  def testGaugeRegistration(self):
    with self.SetUpStatsCollector(
        default_stats_collector.DefaultStatsCollector()
    ):
      metrics.Gauge("gfoo", int)
    self.assertIsNotNone(self.collector.GetMetricMetadata("gfoo"))

  def testEventRegistration(self):
    with self.SetUpStatsCollector(
        default_stats_collector.DefaultStatsCollector()
    ):
      metrics.Event("efoo")
    self.assertIsNotNone(self.collector.GetMetricMetadata("efoo"))

  def testCounterIncrement(self):
    with self.SetUpStatsCollector(
        default_stats_collector.DefaultStatsCollector()
    ):
      counter = metrics.Counter("cfoo", fields=[("bar", str)])
    with self.assertStatsCounterDelta(1, counter, fields=["baz"]):
      counter.Increment(fields=["baz"])

  def testGetValue(self):
    with self.SetUpStatsCollector(
        default_stats_collector.DefaultStatsCollector()
    ):
      counter = metrics.Counter("cfoo", fields=[("bar", str)])
    self.assertEqual(counter.GetValue(["baz"]), 0)
    counter.Increment(fields=["baz"])
    self.assertEqual(counter.GetValue(["baz"]), 1)

  def testGetFields(self):
    with self.SetUpStatsCollector(
        default_stats_collector.DefaultStatsCollector()
    ):
      counter = metrics.Counter("cfoo", fields=[("bar", str)])
    self.assertEmpty(counter.GetFields())
    counter.Increment(fields=["baz"])
    counter.Increment(fields=["bazz"])
    self.assertCountEqual(counter.GetFields(), [("baz",), ("bazz",)])

  def testCountedDecoratorIncrement(self):
    with self.SetUpStatsCollector(
        default_stats_collector.DefaultStatsCollector()
    ):
      counter = metrics.Counter("cfoo", fields=[("bar", str)])

    @counter.Counted(fields=["baz"])
    def Foo():
      pass

    with self.assertStatsCounterDelta(1, counter, fields=["baz"]):
      Foo()

  def testSuccessesCountedDecoratorIncrement(self):
    with self.SetUpStatsCollector(
        default_stats_collector.DefaultStatsCollector()
    ):
      counter = metrics.Counter("cfoo", fields=[("bar", str)])

    @counter.SuccessesCounted(fields=["baz"])
    def Foo():
      pass

    with self.assertStatsCounterDelta(1, counter, fields=["baz"]):
      Foo()

  def testErrorsCountedDecoratorIncrement(self):
    with self.SetUpStatsCollector(
        default_stats_collector.DefaultStatsCollector()
    ):
      counter = metrics.Counter("cfoo", fields=[("bar", str)])

    @counter.ErrorsCounted(fields=["baz"])
    def Foo():
      raise ValueError()

    with self.assertStatsCounterDelta(1, counter, fields=["baz"]):
      with self.assertRaises(ValueError):
        Foo()

  def testSetGaugeValue(self):
    with self.SetUpStatsCollector(
        default_stats_collector.DefaultStatsCollector()
    ):
      gauge = metrics.Gauge("gfoo", int, fields=[("bar", str)])
    with self.assertStatsCounterDelta(42, gauge, fields=["baz"]):
      gauge.SetValue(42, fields=["baz"])

  def testRecordEvent(self):
    with self.SetUpStatsCollector(
        default_stats_collector.DefaultStatsCollector()
    ):
      event = metrics.Event("efoo", fields=[("bar", str)])
    with self.assertStatsCounterDelta(1, event, fields=["baz"]):
      event.RecordEvent(42, fields=["baz"])

  def testTimedDecorator(self):
    with self.SetUpStatsCollector(
        default_stats_collector.DefaultStatsCollector()
    ):
      event = metrics.Event("efoo", fields=[("bar", str)])

    @event.Timed(fields=["baz"])
    def Foo():
      pass

    with self.assertStatsCounterDelta(1, event, fields=["baz"]):
      Foo()

  def testMetricCanBeRegisteredAfterStatsCollectorHasBeenSetUp(self):
    with mock.patch.multiple(
        stats_collector_instance, _metadatas=[], _stats_singleton=None
    ):
      stats_collector_instance.Set(
          default_stats_collector.DefaultStatsCollector()
      )
      counter = metrics.Counter("cfoo")
      counter.Increment(1)


if __name__ == "__main__":
  absltest.main()



#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from typing import Any

from grr_response_core.lib.rdfvalues import protodict as rdf_protodict
from grr_response_proto import jobs_pb2


def ToProtoEmbeddedRDFValue(
    rdf: rdf_protodict.EmbeddedRDFValue,
) -> jobs_pb2.EmbeddedRDFValue:
  return rdf.AsPrimitiveProto()


def ToRDFEmbeddedRDFValue(
    proto: jobs_pb2.EmbeddedRDFValue,
) -> rdf_protodict.EmbeddedRDFValue:
  return rdf_protodict.EmbeddedRDFValue.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoDataBlob(rdf: rdf_protodict.DataBlob) -> jobs_pb2.DataBlob:
  return rdf.AsPrimitiveProto()


def ToRDFDataBlob(proto: jobs_pb2.DataBlob) -> rdf_protodict.DataBlob:
  return rdf_protodict.DataBlob.FromSerializedBytes(proto.SerializeToString())


def ToProtoKeyValue(rdf: rdf_protodict.KeyValue) -> jobs_pb2.KeyValue:
  return rdf.AsPrimitiveProto()


def ToRDFKeyValue(proto: jobs_pb2.KeyValue) -> rdf_protodict.KeyValue:
  return rdf_protodict.KeyValue.FromSerializedBytes(proto.SerializeToString())


def ToProtoDict(rdf: rdf_protodict.Dict) -> jobs_pb2.Dict:
  return rdf.AsPrimitiveProto()


def ToRDFDict(proto: jobs_pb2.Dict) -> rdf_protodict.Dict:
  return rdf_protodict.Dict.FromSerializedBytes(proto.SerializeToString())


def FromProtoDictToNativeDict(proto: jobs_pb2.Dict) -> dict[Any, Any]:
  rdf_dict = ToRDFDict(proto)
  return rdf_dict.ToDict()


def FromNativeDictToProtoDict(dictionary: dict[Any, Any]) -> jobs_pb2.Dict:
  rdf_dict = rdf_protodict.Dict().FromDict(dictionary)
  return ToProtoDict(rdf_dict)


def ToProtoAttributedDict(
    rdf: rdf_protodict.AttributedDict,
) -> jobs_pb2.AttributedDict:
  return rdf.AsPrimitiveProto()


def ToRDFAttributedDict(
    proto: jobs_pb2.AttributedDict,
) -> rdf_protodict.AttributedDict:
  return rdf_protodict.AttributedDict.FromSerializedBytes(
      proto.SerializeToString()
  )


def FromProtoAttributedDictToNativeDict(
    proto: jobs_pb2.AttributedDict,
) -> dict[Any, Any]:
  rdf_dict = ToRDFAttributedDict(proto)
  return rdf_dict.ToDict()


def FromNativeDictToProtoAttributedDict(
    dictionary: dict[Any, Any],
) -> jobs_pb2.AttributedDict:
  rdf_dict = rdf_protodict.AttributedDict().FromDict(dictionary)
  return ToProtoAttributedDict(rdf_dict)


def ToProtoBlobArrayFromBlobArray(
    rdf: rdf_protodict.BlobArray,
) -> jobs_pb2.BlobArray:
  return rdf.AsPrimitiveProto()


def ToRDFBlobArray(proto: jobs_pb2.BlobArray) -> rdf_protodict.BlobArray:
  return rdf_protodict.BlobArray.FromSerializedBytes(proto.SerializeToString())



#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from grr_response_core.lib.rdfvalues import standard as rdf_standard
from grr_response_proto import jobs_pb2
from grr_response_proto import sysinfo_pb2


def ToProtoAuthenticodeSignedData(
    rdf: rdf_standard.AuthenticodeSignedData,
) -> jobs_pb2.AuthenticodeSignedData:
  return rdf.AsPrimitiveProto()


def ToRDFAuthenticodeSignedData(
    proto: jobs_pb2.AuthenticodeSignedData,
) -> rdf_standard.AuthenticodeSignedData:
  return rdf_standard.AuthenticodeSignedData.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoPersistenceFile(
    rdf: rdf_standard.PersistenceFile,
) -> jobs_pb2.PersistenceFile:
  return rdf.AsPrimitiveProto()


def ToRDFPersistenceFile(
    proto: jobs_pb2.PersistenceFile,
) -> rdf_standard.PersistenceFile:
  return rdf_standard.PersistenceFile.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoURI(rdf: rdf_standard.URI) -> sysinfo_pb2.URI:
  return rdf.AsPrimitiveProto()


def ToRDFURI(proto: sysinfo_pb2.URI) -> rdf_standard.URI:
  return rdf_standard.URI.FromSerializedBytes(proto.SerializeToString())



#!/usr/bin/env python
"""A generic serializer for python dictionaries."""

from collections import abc
import logging
from typing import Union, cast

from grr_response_core.lib import rdfvalue
from grr_response_core.lib import serialization
from grr_response_core.lib import utils
from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_proto import jobs_pb2


class EmbeddedRDFValue(rdf_structs.RDFProtoStruct):
  """An object that contains a serialized RDFValue."""

  protobuf = jobs_pb2.EmbeddedRDFValue
  rdf_deps = []

  def __init__(self, initializer=None, payload=None, *args, **kwargs):
    if (
        not payload
        and isinstance(initializer, rdfvalue.RDFValue)
        and not isinstance(initializer, EmbeddedRDFValue)
    ):
      # The initializer is an RDFValue object that we can use as payload.
      payload = initializer
      initializer = None

    super().__init__(initializer=initializer, *args, **kwargs)
    if payload is not None:
      self.payload = payload

  @property
  def payload(self):
    """Extracts and returns the serialized object."""
    try:
      rdf_cls = self.classes.get(self.name)
      if rdf_cls:
        value = rdf_cls.FromSerializedBytes(self.data)

        return value
    except TypeError:
      logging.exception("Error during decoding %s", self)
      return None

  @payload.setter
  def payload(self, payload):
    self.name = payload.__class__.__name__
    self.data = serialization.ToBytes(payload)


class DataBlob(rdf_structs.RDFProtoStruct):
  """Wrapper class for DataBlob protobuf."""

  protobuf = jobs_pb2.DataBlob
  rdf_deps = [
      "BlobArray",  # TODO(user): dependency loop.
      "Dict",  # TODO(user): dependency loop.
      EmbeddedRDFValue,
  ]

  def SetValue(self, value, raise_on_error=True):
    """Receives a value and fills it into a DataBlob.

    Args:
      value: value to set
      raise_on_error: if True, raise if we can't serialize.  If False, set the
        key to an error string.

    Returns:
      self
    Raises:
      TypeError: if the value can't be serialized and raise_on_error is True
    """
    type_mappings = [
        (str, "string"),
        (bytes, "data"),
        (bool, "boolean"),
        (int, "integer"),
        (dict, "dict"),
        (float, "float"),
    ]

    if value is None:
      self.none = "None"

    elif isinstance(value, rdfvalue.RDFValue):
      self.rdf_value.data = value.SerializeToBytes()
      self.rdf_value.name = value.__class__.__name__

    elif isinstance(value, (list, tuple)):
      self.list.content.Extend(
          [DataBlob().SetValue(v, raise_on_error=raise_on_error) for v in value]
      )

    elif isinstance(value, set):
      self.set.content.Extend(
          [DataBlob().SetValue(v, raise_on_error=raise_on_error) for v in value]
      )

    elif isinstance(value, dict):
      self.dict.FromDict(value, raise_on_error=raise_on_error)

    else:
      for type_mapping, member in type_mappings:
        if isinstance(value, type_mapping):
          setattr(self, member, value)

          return self

      message = "Unsupported type for ProtoDict: %s of type %s" % (
          value,
          type(value),
      )
      if raise_on_error:
        raise TypeError(message)

      logging.error(message)
      setattr(self, "string", message)

    return self

  # TODO: Defaulting to ignoring errors is unexpected and
  #  problematic.
  def GetValue(self, ignore_error=True):
    """Extracts and returns a single value from a DataBlob."""
    if self.HasField("none"):
      return None

    field_names = [
        "integer",
        "string",
        "data",
        "boolean",
        "list",
        "dict",
        "rdf_value",
        "float",
        "set",
    ]

    values = [getattr(self, x) for x in field_names if self.HasField(x)]

    if len(values) != 1:
      return None

    if self.HasField("boolean"):
      return bool(values[0])

    # Unpack RDFValues.
    if self.HasField("rdf_value"):
      try:
        rdf_class = rdfvalue.RDFValue.classes[self.rdf_value.name]
        return serialization.FromBytes(rdf_class, self.rdf_value.data)
      except (ValueError, KeyError) as e:
        if ignore_error:
          logging.exception("Error during GetValue() of %s", self)
          return e

        raise

    elif self.HasField("list"):
      return [x.GetValue() for x in self.list.content]

    elif self.HasField("set"):
      return set([x.GetValue() for x in self.set.content])

    else:
      return values[0]


class KeyValue(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.KeyValue
  rdf_deps = [
      DataBlob,
  ]


class Dict(rdf_structs.RDFProtoStruct):
  """A high level interface for protobuf Dict objects.

  This effectively converts from a dict to a proto and back.
  The dict may contain strings (python unicode objects), int64,
  or binary blobs (python string objects) as keys and values.
  """

  protobuf = jobs_pb2.Dict
  rdf_deps = [
      KeyValue,
  ]

  _values = None

  def __init__(self, initializer=None, **kwargs):
    super().__init__(initializer=None)

    self.dat: Union[list[KeyValue], rdf_structs.RepeatedFieldHelper] = None

    # Support initializing from a mapping
    if isinstance(initializer, dict):
      self.FromDict(initializer)

    # Can be initialized from kwargs (like a dict).
    elif initializer is None:
      self.FromDict(kwargs)

    # Initialize from another Dict.
    elif isinstance(initializer, Dict):
      self.FromDict(initializer.ToDict())

    else:
      raise rdfvalue.InitializeError("Invalid initializer for ProtoDict.")

  def ToDict(self):

    def _Convert(obj):
      if isinstance(obj, Dict):
        return {key: _Convert(value) for key, value in obj.items()}
      if isinstance(obj, list):
        return [_Convert(item) for item in obj]
      return obj

    return _Convert(self)

  def FromDict(self, dictionary, raise_on_error=True):
    # First clear and then set the dictionary.
    self._values = {}
    for key, value in dictionary.items():
      self._values[key] = KeyValue(
          k=DataBlob().SetValue(key, raise_on_error=raise_on_error),
          v=DataBlob().SetValue(value, raise_on_error=raise_on_error),
      )
    self.dat = self._values.values()  # pytype: disable=annotation-type-mismatch
    return self

  def __getitem__(self, key):
    return self._values[key].v.GetValue()

  def __contains__(self, key):
    return key in self._values

  # TODO: This implementation is flawed. It returns a new instance
  # on each invocation, effectively preventing changes to mutable
  # datastructures, e.g. `dct["key"] = []; dct["key"].append(5)`.
  def GetItem(self, key, default=None):
    if key in self._values:
      return self._values[key].v.GetValue()
    return default

  def Items(self):
    for x in self._values.values():
      yield x.k.GetValue(), x.v.GetValue()

  def Values(self):
    for x in self._values.values():
      yield x.v.GetValue()

  def Keys(self):
    for x in self._values.values():
      yield x.k.GetValue()

  get = utils.Proxy("GetItem")
  items = utils.Proxy("Items")
  keys = utils.Proxy("Keys")
  values = utils.Proxy("Values")

  def __delitem__(self, key):
    # TODO(user):pytype: assigning "dirty" here is a hack. The assumption
    # that self.dat is RepeatedFieldHelper may not hold. For some reason the
    # type checker doesn not respect the isinstance check below and explicit
    # cast is required.
    if not isinstance(self.dat, rdf_structs.RepeatedFieldHelper):
      raise TypeError("self.dat has an unexpected type %s" % self.dat.__class__)
    cast(rdf_structs.RepeatedFieldHelper, self.dat).dirty = True
    del self._values[key]

  def __len__(self):
    return len(self._values)

  def SetItem(self, key, value, raise_on_error=True):
    """Alternative to __setitem__ that can ignore errors.

    Sometimes we want to serialize a structure that contains some simple
    objects, and some that can't be serialized.  This method gives the caller a
    way to specify that they don't care about values that can't be
    serialized.

    Args:
      key: dict key
      value: dict value
      raise_on_error: if True, raise if we can't serialize.  If False, set the
        key to an error string.
    """
    # TODO(user):pytype: assigning "dirty" here is a hack. The assumption
    # that self.dat is RepeatedFieldHelper may not hold. For some reason the
    # type checker doesn not respect the isinstance check below and explicit
    # cast is required.
    if not isinstance(self.dat, rdf_structs.RepeatedFieldHelper):
      raise TypeError("self.dat has an unexpected type %s" % self.dat.__class__)
    cast(rdf_structs.RepeatedFieldHelper, self.dat).dirty = True
    self._values[key] = KeyValue(
        k=DataBlob().SetValue(key, raise_on_error=raise_on_error),
        v=DataBlob().SetValue(value, raise_on_error=raise_on_error),
    )

  def __setitem__(self, key, value):
    # TODO(user):pytype: assigning "dirty" here is a hack. The assumption
    # that self.dat is RepeatedFieldHelper may not hold. For some reason the
    # type checker doesn not respect the isinstance check below and explicit
    # cast is required.
    if not isinstance(self.dat, rdf_structs.RepeatedFieldHelper):
      raise TypeError("self.dat has an unexpected type %s" % self.dat.__class__)
    cast(rdf_structs.RepeatedFieldHelper, self.dat).dirty = True
    self._values[key] = KeyValue(
        k=DataBlob().SetValue(key), v=DataBlob().SetValue(value)
    )

  def __iter__(self):
    for x in self._values.values():
      yield x.k.GetValue()

  # Required, because in Python 3 overriding `__eq__` nullifies `__hash__`.
  __hash__ = rdf_structs.RDFProtoStruct.__hash__

  def __eq__(self, other):
    if isinstance(other, dict):
      return self.ToDict() == other
    elif isinstance(other, Dict):
      return self.ToDict() == other.ToDict()
    else:
      return False

  def GetRawData(self):
    self.dat = self._values.values()  # pytype: disable=annotation-type-mismatch
    return super().GetRawData()

  def _CopyRawData(self):
    self.dat = self._values.values()  # pytype: disable=annotation-type-mismatch
    return super()._CopyRawData()

  def SetRawData(self, raw_data):
    super().SetRawData(raw_data)
    self._values = {}
    for d in self.dat:
      self._values[d.k.GetValue()] = d

  def SerializeToBytes(self):
    self.dat = self._values.values()  # pytype: disable=annotation-type-mismatch
    return super().SerializeToBytes()

  def __str__(self) -> str:
    return str(self.ToDict())


class AttributedDict(Dict):
  """A Dict that supports attribute indexing."""

  protobuf = jobs_pb2.AttributedDict
  rdf_deps = [
      KeyValue,
  ]

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self._StringifyKeys()

  def SetRawData(self, raw_data):
    super().SetRawData(raw_data)
    self._StringifyKeys()

  def __getattr__(self, item):
    # Pickle is checking for the presence of overrides for various builtins.
    # Without this check we swallow the error and return None, which confuses
    # pickle protocol version 2.
    if item.startswith("__"):
      raise AttributeError()
    return self.GetItem(item)

  def __setattr__(self, item, value):
    # Existing class or instance members are assigned to normally.
    if hasattr(self.__class__, item) or item in self.__dict__:
      object.__setattr__(self, item, value)
    else:
      self.SetItem(item, value)

  def __setitem__(self, key, value):
    # TODO: This behavior should be removed once migration is done.
    if isinstance(key, bytes):
      key = key.decode("utf-8")

    if isinstance(key, str):
      return super().__setitem__(key, value)

    raise TypeError("Non-string key: {!r}".format(key))

  def __getitem__(self, key):
    # TODO: This behavior should be removed once migration is done.
    if isinstance(key, bytes):
      key = key.decode("utf-8")

    if isinstance(key, str):
      return super().__getitem__(key)

    raise TypeError("Non-string key: {!r}".format(key))

  # TODO: This behavior should be removed once migration is done.
  # Because of Python 3 migration and incompatibilities between Pythons in how
  # attributed dicts are serialized, we are forced to have this dirty hack in
  # here. Once migration is done and we are sure that the old serialized data
  # that is stored in the database does not need to be read anymore, these can
  # be removed.
  def _StringifyKeys(self):
    """Turns byte string keys into unicode string keys."""
    byte_keys = set()

    for key in self._values.keys():
      if isinstance(key, bytes):
        byte_keys.add(key)
      elif not isinstance(key, str):
        raise TypeError("Non-string key: {!r}".format(key))

    for byte_key in byte_keys:
      value = self._values[byte_key].v
      del self._values[byte_key]

      string_key = byte_key.decode("utf-8")

      entry = KeyValue(k=DataBlob().SetValue(string_key), v=value)
      self._values[string_key] = entry

    # If we made any changes, update `self.dat` (whatever it is, but other
    # methods seem to do the same in case of modifying the internal state).
    if byte_keys:
      self.dat = list(self._values.values())


class BlobArray(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.BlobArray
  rdf_deps = [
      DataBlob,
  ]


# TODO(user):pytype: Mapping is likely using abc.ABCMeta that provides a
# "register" method. Type checker doesn't see this, unfortunately.
abc.Mapping.register(Dict)  # pytype: disable=attribute-error



#!/usr/bin/env python
"""Standard RDFValues."""

import re
from urllib import parse as urlparse

from grr_response_core.lib import config_lib
from grr_response_core.lib import rdfvalue
from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_core.lib.util import precondition
from grr_response_proto import jobs_pb2
from grr_response_proto import sysinfo_pb2


class RegularExpression(rdfvalue.RDFString):
  """A semantic regular expression."""

  context_help_url = (
      "investigating-with-grr/flows/"
      "literal-and-regex-matching.html#regex-matches"
  )

  def __init__(self, initializer=None):
    super().__init__(initializer=initializer)
    # Try compiling the pattern right away to fail fast for pattern errors.
    self._Regex()

  def _Regex(self):
    return re.compile(self._value, flags=re.I | re.S | re.M)

  def Search(self, text):
    """Search the text for our value."""
    if isinstance(text, rdfvalue.RDFString):
      text = str(text)

    return self._Regex().search(text)

  def Match(self, text):
    if isinstance(text, rdfvalue.RDFString):
      text = str(text)

    return self._Regex().match(text)

  def FindIter(self, text):
    if isinstance(text, rdfvalue.RDFString):
      text = str(text)

    return self._Regex().finditer(text)


class LiteralExpression(rdfvalue.RDFBytes):
  """A RDFBytes literal for use in GrepSpec."""

  context_help_url = (
      "investigating-with-grr/flows/"
      "literal-and-regex-matching.html#literal-matches"
  )


class EmailAddress(rdfvalue.RDFString):
  """An email address must be well formed."""

  _EMAIL_REGEX = re.compile(r"[^@]+@([^@]+)$")

  def __init__(self, initializer=None):
    super().__init__(initializer)

    self._match = self._EMAIL_REGEX.match(self._value)
    if self._value and not self._match:
      raise ValueError("Email address %r not well formed." % self._value)


class DomainEmailAddress(EmailAddress):
  """A more restricted email address may only address the domain."""

  def __init__(self, initializer=None):
    super().__init__(initializer)

    # TODO(user): dependency loop with
    # core/grr_response_core/grr/config/client.py.
    # pylint: disable=protected-access
    domain = config_lib._CONFIG["Logging.domain"]
    # pylint: enable=protected-access
    if self._value and domain and self._match.group(1) != domain:
      raise ValueError(
          "Email address '%s' does not belong to the configured domain '%s'"
          % (self._match.group(1), domain)
      )


class AuthenticodeSignedData(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.AuthenticodeSignedData


class PersistenceFile(rdf_structs.RDFProtoStruct):
  protobuf = jobs_pb2.PersistenceFile
  rdf_deps = [
      "PathSpec",  # TODO(user): dependency loop.
      rdfvalue.RDFURN,
  ]


class URI(rdf_structs.RDFProtoStruct):
  """Represets a URI with its individual components separated."""

  protobuf = sysinfo_pb2.URI

  def __init__(self, initializer=None, **kwargs):
    if not isinstance(initializer, urlparse.ParseResult):
      super().__init__(initializer, **kwargs)
      return

    super().__init__()

    if initializer.scheme:
      self.transport = initializer.scheme
    if initializer.netloc:
      self.host = initializer.netloc
    if initializer.path:
      self.path = initializer.path
    if initializer.query:
      self.query = initializer.query
    if initializer.fragment:
      self.fragment = initializer.fragment

  @classmethod
  def FromSerializedBytes(cls, value: bytes):
    precondition.AssertType(value, bytes)
    return cls(urlparse.urlparse(value.decode("utf-8")))

  @classmethod
  def FromHumanReadable(cls, value: str):
    precondition.AssertType(value, str)
    return cls(urlparse.urlparse(value))

  def SerializeToBytes(self) -> bytes:
    return self.SerializeToHumanReadable().encode("utf-8")

  def SerializeToHumanReadable(self) -> str:
    parts = (self.transport, self.host, self.path, self.query, self.fragment)
    return urlparse.urlunsplit(parts)  # pytype: disable=bad-return-type



#!/usr/bin/env python
"""Proto functions used with Yara."""

from grr_response_proto import flows_pb2


def CreateYaraProcessScanRequest(
    include_errors_in_results: flows_pb2.YaraProcessScanRequest.ErrorPolicy = flows_pb2.YaraProcessScanRequest.ErrorPolicy.NO_ERRORS,
    include_misses_in_results: bool = False,
    ignore_grr_process: bool = True,
    chunk_size: int = 100 * 1024 * 1024,
    overlap_size: int = 10 * 1024 * 1024,
    skip_special_regions: bool = False,
    skip_mapped_files: bool = True,
    skip_shared_regions: bool = False,
    skip_executable_regions: bool = False,
    skip_readonly_regions: bool = False,
    dump_process_on_match: bool = False,
    process_dump_size_limit: int = 0,
    context_window: int = 50,
) -> flows_pb2.YaraProcessScanRequest:
  """Creates a YaraProcessScanRequest with default values."""

  return flows_pb2.YaraProcessScanRequest(
      include_errors_in_results=include_errors_in_results,
      include_misses_in_results=include_misses_in_results,
      ignore_grr_process=ignore_grr_process,
      chunk_size=chunk_size,
      overlap_size=overlap_size,
      skip_special_regions=skip_special_regions,
      skip_mapped_files=skip_mapped_files,
      skip_shared_regions=skip_shared_regions,
      skip_executable_regions=skip_executable_regions,
      skip_readonly_regions=skip_readonly_regions,
      dump_process_on_match=dump_process_on_match,
      process_dump_size_limit=process_dump_size_limit,
      context_window=context_window,
  )



#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from grr_response_core.lib.rdfvalues import file_finder as rdf_file_finder
from grr_response_proto import flows_pb2


def ToProtoFileFinderModificationTimeCondition(
    rdf: rdf_file_finder.FileFinderModificationTimeCondition,
) -> flows_pb2.FileFinderModificationTimeCondition:
  return rdf.AsPrimitiveProto()


def ToRDFFileFinderModificationTimeCondition(
    proto: flows_pb2.FileFinderModificationTimeCondition,
) -> rdf_file_finder.FileFinderModificationTimeCondition:
  return (
      rdf_file_finder.FileFinderModificationTimeCondition.FromSerializedBytes(
          proto.SerializeToString()
      )
  )


def ToProtoFileFinderAccessTimeCondition(
    rdf: rdf_file_finder.FileFinderAccessTimeCondition,
) -> flows_pb2.FileFinderAccessTimeCondition:
  return rdf.AsPrimitiveProto()


def ToRDFFileFinderAccessTimeCondition(
    proto: flows_pb2.FileFinderAccessTimeCondition,
) -> rdf_file_finder.FileFinderAccessTimeCondition:
  return rdf_file_finder.FileFinderAccessTimeCondition.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoFileFinderInodeChangeTimeCondition(
    rdf: rdf_file_finder.FileFinderInodeChangeTimeCondition,
) -> flows_pb2.FileFinderInodeChangeTimeCondition:
  return rdf.AsPrimitiveProto()


def ToRDFFileFinderInodeChangeTimeCondition(
    proto: flows_pb2.FileFinderInodeChangeTimeCondition,
) -> rdf_file_finder.FileFinderInodeChangeTimeCondition:
  return rdf_file_finder.FileFinderInodeChangeTimeCondition.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoFileFinderSizeCondition(
    rdf: rdf_file_finder.FileFinderSizeCondition,
) -> flows_pb2.FileFinderSizeCondition:
  return rdf.AsPrimitiveProto()


def ToRDFFileFinderSizeCondition(
    proto: flows_pb2.FileFinderSizeCondition,
) -> rdf_file_finder.FileFinderSizeCondition:
  return rdf_file_finder.FileFinderSizeCondition.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoFileFinderExtFlagsCondition(
    rdf: rdf_file_finder.FileFinderExtFlagsCondition,
) -> flows_pb2.FileFinderExtFlagsCondition:
  return rdf.AsPrimitiveProto()


def ToRDFFileFinderExtFlagsCondition(
    proto: flows_pb2.FileFinderExtFlagsCondition,
) -> rdf_file_finder.FileFinderExtFlagsCondition:
  return rdf_file_finder.FileFinderExtFlagsCondition.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoFileFinderContentsRegexMatchCondition(
    rdf: rdf_file_finder.FileFinderContentsRegexMatchCondition,
) -> flows_pb2.FileFinderContentsRegexMatchCondition:
  return rdf.AsPrimitiveProto()


def ToRDFFileFinderContentsRegexMatchCondition(
    proto: flows_pb2.FileFinderContentsRegexMatchCondition,
) -> rdf_file_finder.FileFinderContentsRegexMatchCondition:
  return (
      rdf_file_finder.FileFinderContentsRegexMatchCondition.FromSerializedBytes(
          proto.SerializeToString()
      )
  )


def ToProtoFileFinderContentsLiteralMatchCondition(
    rdf: rdf_file_finder.FileFinderContentsLiteralMatchCondition,
) -> flows_pb2.FileFinderContentsLiteralMatchCondition:
  return rdf.AsPrimitiveProto()


def ToRDFFileFinderContentsLiteralMatchCondition(
    proto: flows_pb2.FileFinderContentsLiteralMatchCondition,
) -> rdf_file_finder.FileFinderContentsLiteralMatchCondition:
  return rdf_file_finder.FileFinderContentsLiteralMatchCondition.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoFileFinderCondition(
    rdf: rdf_file_finder.FileFinderCondition,
) -> flows_pb2.FileFinderCondition:
  return rdf.AsPrimitiveProto()


def ToRDFFileFinderCondition(
    proto: flows_pb2.FileFinderCondition,
) -> rdf_file_finder.FileFinderCondition:
  return rdf_file_finder.FileFinderCondition.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoFileFinderStatActionOptions(
    rdf: rdf_file_finder.FileFinderStatActionOptions,
) -> flows_pb2.FileFinderStatActionOptions:
  return rdf.AsPrimitiveProto()


def ToRDFFileFinderStatActionOptions(
    proto: flows_pb2.FileFinderStatActionOptions,
) -> rdf_file_finder.FileFinderStatActionOptions:
  return rdf_file_finder.FileFinderStatActionOptions.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoFileFinderHashActionOptions(
    rdf: rdf_file_finder.FileFinderHashActionOptions,
) -> flows_pb2.FileFinderHashActionOptions:
  return rdf.AsPrimitiveProto()


def ToRDFFileFinderHashActionOptions(
    proto: flows_pb2.FileFinderHashActionOptions,
) -> rdf_file_finder.FileFinderHashActionOptions:
  return rdf_file_finder.FileFinderHashActionOptions.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoFileFinderDownloadActionOptions(
    rdf: rdf_file_finder.FileFinderDownloadActionOptions,
) -> flows_pb2.FileFinderDownloadActionOptions:
  return rdf.AsPrimitiveProto()


def ToRDFFileFinderDownloadActionOptions(
    proto: flows_pb2.FileFinderDownloadActionOptions,
) -> rdf_file_finder.FileFinderDownloadActionOptions:
  return rdf_file_finder.FileFinderDownloadActionOptions.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoFileFinderAction(
    rdf: rdf_file_finder.FileFinderAction,
) -> flows_pb2.FileFinderAction:
  return rdf.AsPrimitiveProto()


def ToRDFFileFinderAction(
    proto: flows_pb2.FileFinderAction,
) -> rdf_file_finder.FileFinderAction:
  return rdf_file_finder.FileFinderAction.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoFileFinderArgs(
    rdf: rdf_file_finder.FileFinderArgs,
) -> flows_pb2.FileFinderArgs:
  return rdf.AsPrimitiveProto()


def ToRDFFileFinderArgs(
    proto: flows_pb2.FileFinderArgs,
) -> rdf_file_finder.FileFinderArgs:
  return rdf_file_finder.FileFinderArgs.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoFileFinderResult(
    rdf: rdf_file_finder.FileFinderResult,
) -> flows_pb2.FileFinderResult:
  return rdf.AsPrimitiveProto()


def ToRDFFileFinderResult(
    proto: flows_pb2.FileFinderResult,
) -> rdf_file_finder.FileFinderResult:
  return rdf_file_finder.FileFinderResult.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoCollectFilesByKnownPathArgs(
    rdf: rdf_file_finder.CollectFilesByKnownPathArgs,
) -> flows_pb2.CollectFilesByKnownPathArgs:
  return rdf.AsPrimitiveProto()


def ToRDFCollectFilesByKnownPathArgs(
    proto: flows_pb2.CollectFilesByKnownPathArgs,
) -> rdf_file_finder.CollectFilesByKnownPathArgs:
  return rdf_file_finder.CollectFilesByKnownPathArgs.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoCollectFilesByKnownPathResult(
    rdf: rdf_file_finder.CollectFilesByKnownPathResult,
) -> flows_pb2.CollectFilesByKnownPathResult:
  return rdf.AsPrimitiveProto()


def ToRDFCollectFilesByKnownPathResult(
    proto: flows_pb2.CollectFilesByKnownPathResult,
) -> rdf_file_finder.CollectFilesByKnownPathResult:
  return rdf_file_finder.CollectFilesByKnownPathResult.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoCollectFilesByKnownPathProgress(
    rdf: rdf_file_finder.CollectFilesByKnownPathProgress,
) -> flows_pb2.CollectFilesByKnownPathProgress:
  return rdf.AsPrimitiveProto()


def ToRDFCollectFilesByKnownPathProgress(
    proto: flows_pb2.CollectFilesByKnownPathProgress,
) -> rdf_file_finder.CollectFilesByKnownPathProgress:
  return rdf_file_finder.CollectFilesByKnownPathProgress.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoCollectMultipleFilesArgs(
    rdf: rdf_file_finder.CollectMultipleFilesArgs,
) -> flows_pb2.CollectMultipleFilesArgs:
  return rdf.AsPrimitiveProto()


def ToRDFCollectMultipleFilesArgs(
    proto: flows_pb2.CollectMultipleFilesArgs,
) -> rdf_file_finder.CollectMultipleFilesArgs:
  return rdf_file_finder.CollectMultipleFilesArgs.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoCollectMultipleFilesResult(
    rdf: rdf_file_finder.CollectMultipleFilesResult,
) -> flows_pb2.CollectMultipleFilesResult:
  return rdf.AsPrimitiveProto()


def ToRDFCollectMultipleFilesResult(
    proto: flows_pb2.CollectMultipleFilesResult,
) -> rdf_file_finder.CollectMultipleFilesResult:
  return rdf_file_finder.CollectMultipleFilesResult.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoCollectMultipleFilesProgress(
    rdf: rdf_file_finder.CollectMultipleFilesProgress,
) -> flows_pb2.CollectMultipleFilesProgress:
  return rdf.AsPrimitiveProto()


def ToRDFCollectMultipleFilesProgress(
    proto: flows_pb2.CollectMultipleFilesProgress,
) -> rdf_file_finder.CollectMultipleFilesProgress:
  return rdf_file_finder.CollectMultipleFilesProgress.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoStatMultipleFilesArgs(
    rdf: rdf_file_finder.StatMultipleFilesArgs,
) -> flows_pb2.StatMultipleFilesArgs:
  return rdf.AsPrimitiveProto()


def ToRDFStatMultipleFilesArgs(
    proto: flows_pb2.StatMultipleFilesArgs,
) -> rdf_file_finder.StatMultipleFilesArgs:
  return rdf_file_finder.StatMultipleFilesArgs.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoHashMultipleFilesArgs(
    rdf: rdf_file_finder.HashMultipleFilesArgs,
) -> flows_pb2.HashMultipleFilesArgs:
  return rdf.AsPrimitiveProto()


def ToRDFHashMultipleFilesArgs(
    proto: flows_pb2.HashMultipleFilesArgs,
) -> rdf_file_finder.HashMultipleFilesArgs:
  return rdf_file_finder.HashMultipleFilesArgs.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoHashMultipleFilesProgress(
    rdf: rdf_file_finder.HashMultipleFilesProgress,
) -> flows_pb2.HashMultipleFilesProgress:
  return rdf.AsPrimitiveProto()


def ToRDFHashMultipleFilesProgress(
    proto: flows_pb2.HashMultipleFilesProgress,
) -> rdf_file_finder.HashMultipleFilesProgress:
  return rdf_file_finder.HashMultipleFilesProgress.FromSerializedBytes(
      proto.SerializeToString()
  )



#!/usr/bin/env python
"""GRR Rapid Response Framework."""


import configparser
import os

from grr_response_core.lib import package


def VersionPath():
  """Returns a path to version.ini."""

  # Try to get a version.ini. It should be in the resources if the code
  # was packed with "pip sdist". It will be 2 levels up from grr_response_core
  # if the code was installed via "pip install -e".
  version_ini = (
      package.ResourcePath("grr-response-core", "version.ini") or
      package.ResourcePath("grr-response-core", "../../version.ini"))

  if not os.path.exists(version_ini):
    raise RuntimeError("Can't find version.ini at %s" % version_ini)

  return version_ini


def Version():
  """Return a dict with GRR version information."""

  version_ini = VersionPath()

  config = configparser.ConfigParser()
  config.read(version_ini)
  return dict(
      packageversion=config.get("Version", "packageversion"),
      major=config.getint("Version", "major"),
      minor=config.getint("Version", "minor"),
      revision=config.getint("Version", "revision"),
      release=config.getint("Version", "release"))



#!/usr/bin/env python
"""Pathspecs are methods of specifying the path on the client.

The GRR client has a number of drivers to virtualize access to different objects
to create a Virtual File System (VFS) abstraction. These are called 'VFS
Handlers' and they provide typical file-like operations (e.g. read, seek, tell
and stat).  It is possible to recursively apply different drivers in the correct
order to arrive at a certain file like object. In order to specify how drivers
should be applied we use 'Path Specifications' or pathspec.

Each VFS handler is constructed from a previous handler and a pathspec. The
pathspec is just a collection of arguments which make sense to the specific VFS
handler. The type of the handler is carried by the pathtype parameter.

On the server the PathSpec is represented as a PathSpec object, and stored
as an attribute of the AFF4 object. This module defines this abstraction.
"""

from collections.abc import Iterable, Iterator, Sequence
import itertools
import posixpath
import re
from typing import Optional

from grr_response_core.lib import artifact_utils
from grr_response_core.lib import rdfvalue
from grr_response_core.lib import utils
from grr_response_core.lib.rdfvalues import standard as rdf_standard
from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_proto import flows_pb2
from grr_response_proto import jobs_pb2
from grr_response_proto import knowledge_base_pb2


class PathSpec(rdf_structs.RDFProtoStruct):
  """A path specification.

  The pathspec protobuf is a recursive protobuf which contains components. This
  class makes it easier to manipulate these structures by providing useful
  helpers.
  """

  protobuf = jobs_pb2.PathSpec
  rdf_deps = [
      rdfvalue.ByteSize,
      "PathSpec",  # TODO(user): recursive definition.
  ]

  @classmethod
  def OS(cls, **kwargs):
    return cls(pathtype=PathSpec.PathType.OS, **kwargs)

  @classmethod
  def TSK(cls, **kwargs):
    return cls(pathtype=PathSpec.PathType.TSK, **kwargs)

  @classmethod
  def NTFS(cls, **kwargs):
    return cls(pathtype=PathSpec.PathType.NTFS, **kwargs)

  @classmethod
  def Registry(cls, **kwargs):
    return cls(pathtype=PathSpec.PathType.REGISTRY, **kwargs)

  @classmethod
  def Temp(cls, **kwargs):
    return cls(pathtype=PathSpec.PathType.TMPFILE, **kwargs)

  def __len__(self):
    """Return the total number of path components."""
    i = -1
    # TODO(user):pytype: type checker doesn't treat self as iterable.
    for i, _ in enumerate(self):  # pytype: disable=wrong-arg-types
      pass

    return i + 1

  def __getitem__(self, item):
    # TODO(user):pytype: type checker doesn't treat self as iterable.
    for i, element in enumerate(self):  # pytype: disable=wrong-arg-types
      if i == item:
        return element

    raise IndexError("Pathspec index (%s) out of range" % item)

  def __iter__(self):
    """Only iterate over all components from the current pointer."""
    element = self

    while element.HasField("pathtype"):
      yield element

      if element.HasField("nested_path"):
        element = element.nested_path
      else:
        break

  def Insert(self, index, rdfpathspec=None, **kwarg):
    """Insert a single component at index."""
    if rdfpathspec is None:
      rdfpathspec = self.__class__(**kwarg)

    if index == 0:
      # Copy ourselves to a temp copy.
      nested_proto = self.__class__()
      nested_proto.SetRawData(self.GetRawData())

      # Replace ourselves with the new object.
      self.SetRawData(rdfpathspec.GetRawData())

      # Append the temp copy to the end.
      self.last.nested_path = nested_proto
    else:
      previous = self[index - 1]
      rdfpathspec.last.nested_path = previous.nested_path
      previous.nested_path = rdfpathspec

  def Append(self, component=None, **kwarg):
    """Append a new pathspec component to this pathspec."""
    if component is None:
      component = self.__class__(**kwarg)

    if self.HasField("pathtype"):
      self.last.nested_path = component
    else:
      for k, v in kwarg.items():
        setattr(self, k, v)

      self.SetRawData(component.GetRawData())

    return self

  def CollapsePath(self):
    return utils.JoinPath(*[x.path for x in self])

  def Pop(self, index=0):
    """Removes and returns the pathspec at the specified index."""
    if index < 0:
      index += len(self)

    if index == 0:
      result = self.__class__()
      result.SetRawData(self.GetRawData())

      self.SetRawData(self.nested_path.GetRawData())

    else:
      # Get the raw protobufs for the previous member.
      previous = self[index - 1]

      result = previous.nested_path

      # Manipulate the previous members protobuf to patch the next component in.
      previous.nested_path = result.nested_path

    result.nested_path = None

    return result

  @property
  def first(self):
    return self

  @property
  def last(self):
    if self.HasField("pathtype") and self.pathtype != self.PathType.UNSET:
      # TODO(user):pytype: type checker doesn't treat self as iterable.
      return list(self)[-1]  # pytype: disable=wrong-arg-types

    return self

  def Dirname(self):
    """Get a new copied object with only the directory path."""
    result = self.Copy()

    while 1:
      last_directory = posixpath.dirname(result.last.path)
      if last_directory != "/" or len(result) <= 1:
        result.last.path = last_directory
        # Make sure to clear the inode information.
        result.last.inode = None

        break

      result.Pop(-1)

    return result

  def Basename(self):
    # TODO(user):pytype: type checker doesn't treat self as reversible.
    for component in reversed(self):  # pytype: disable=wrong-arg-types
      basename = posixpath.basename(component.path)
      if basename:
        return basename

    return ""

  def Validate(self):
    if not self.HasField("pathtype") or self.pathtype == self.PathType.UNSET:
      raise ValueError("No path type set in PathSpec.")

  AFF4_PREFIXES = {
      0: "/fs/os",  # PathSpec.PathType.OS
      1: "/fs/tsk",  # PathSpec.PathType.TSK
      2: "/registry",  # PathSpec.PathType.REGISTRY
      4: "/temp",  # PathSpec.PathType.TMPFILE
      5: "/fs/ntfs",  # PathSpec.PathType.NTFS
  }

  def AFF4Path(self, client_urn):
    """Returns the AFF4 URN this pathspec will be stored under.

    Args:
      client_urn: A ClientURN.

    Returns:
      A urn that corresponds to this pathspec.

    Raises:
      ValueError: If pathspec is not of the correct type.
    """
    # If the first level is OS and the second level is TSK its probably a mount
    # point resolution. We map it into the tsk branch. For example if we get:
    # path: \\\\.\\Volume{1234}\\
    # pathtype: OS
    # mount_point: /c:/
    # nested_path {
    #    path: /windows/
    #    pathtype: TSK
    # }
    # We map this to aff4://client_id/fs/tsk/\\\\.\\Volume{1234}\\/windows/
    # (The same applies for NTFS)

    if not self.HasField("pathtype"):
      raise ValueError(
          "Can't determine AFF4 path without a valid pathtype for {}.".format(
              self
          )
      )

    first_component = self[0]
    dev = first_component.path
    if first_component.HasField("offset"):
      # We divide here just to get prettier numbers in the GUI
      dev += ":{}".format(first_component.offset // 512)

    if (
        len(self) > 1
        and first_component.pathtype == PathSpec.PathType.OS
        and self[1].pathtype in (PathSpec.PathType.TSK, PathSpec.PathType.NTFS)
    ):
      result = [self.AFF4_PREFIXES[self[1].pathtype], dev]

      # Skip the top level pathspec.
      start = 1
    else:
      # For now just map the top level prefix based on the first pathtype
      result = [self.AFF4_PREFIXES[first_component.pathtype]]
      start = 0

    for p in self[start]:
      component = p.path

      # The following encode different pathspec properties into the AFF4 path in
      # such a way that unique files on the client are mapped to unique URNs in
      # the AFF4 space. Note that this transformation does not need to be
      # reversible since we always use the PathSpec when accessing files on the
      # client.
      if p.HasField("offset"):
        component += ":{}".format(p.offset // 512)

      # Support ADS names.
      if p.HasField("stream_name"):
        component += ":" + p.stream_name

      result.append(component)

    return client_urn.Add("/".join(result))


def _Unique(iterable: Iterable[str]) -> Sequence[str]:
  """Returns a list of unique values in preserved order."""
  return list(dict.fromkeys(iterable))


class GlobComponentExplanation(rdf_structs.RDFProtoStruct):
  """A sub-part of a GlobExpression with examples."""

  protobuf = flows_pb2.GlobComponentExplanation


# Grouping pattern: e.g. {test.exe,foo.doc,bar.txt}
GROUPING_PATTERN = re.compile("{([^}]+,[^}]+)}")

_VAR_PATTERN = re.compile("(" + "|".join([r"%%\w+%%", r"%%\w+\.\w+%%"]) + ")")

_REGEX_SPLIT_PATTERN = re.compile(
    "(" + "|".join(["{[^}]+,[^}]+}", "\\?", "\\*\\*\\/?", "\\*"]) + ")"
)

_COMPONENT_SPLIT_PATTERN = re.compile(
    "("
    + "|".join([
        r"{[^}]+,[^}]+}",
        r"\?",
        r"\*\*\d*/?",
        r"\*",
        r"%%\w+%%",
        r"%%\w+\.\w+%%",
    ])
    + ")"
)


class GlobExpression(rdfvalue.RDFString):
  """A glob expression for a client path.

  A glob expression represents a set of regular expressions which match files on
  the client. The Glob expression supports the following expansions:

  1) Client attribute expansions are surrounded with %% characters. They will be
     expanded from the client AFF4 object.

  2) Groupings are collections of alternates. e.g. {foo.exe,bar.sys}
  3) Wild cards like * and ?
  """

  context_help_url = "investigating-with-grr/flows/specifying-file-paths.html"

  RECURSION_REGEX = re.compile(r"\*\*(\d*)")

  def Validate(self):
    """GlobExpression is valid."""
    if len(self.RECURSION_REGEX.findall(self._value)) > 1:
      raise ValueError("Only one ** is permitted per path: %s." % self._value)

  def Interpolate(
      self,
      knowledge_base: Optional[knowledge_base_pb2.KnowledgeBase] = None,
  ) -> Iterator[str]:
    interpolation = artifact_utils.KnowledgeBaseInterpolation(
        pattern=self._value,
        kb=knowledge_base or knowledge_base_pb2.KnowledgeBase(),
    )

    for pattern in interpolation.results:
      # Normalize the component path (this allows us to resolve ../
      # sequences).
      pattern = utils.NormalizePath(pattern.replace("\\", "/"))

      for p in self.InterpolateGrouping(pattern):
        yield p

  def InterpolateGrouping(
      self,
      pattern: str,
  ) -> Iterator[str]:
    """Interpolate inline globbing groups."""
    components = []
    offset = 0
    for match in GROUPING_PATTERN.finditer(pattern):
      components.append([pattern[offset : match.start()]])

      # Expand the attribute into the set of possibilities:
      alternatives = match.group(1).split(",")
      components.append(_Unique(alternatives))
      offset = match.end()

    components.append([pattern[offset:]])
    # Now calculate the cartesian products of all these sets to form all
    # strings.
    for vector in itertools.product(*components):
      yield "".join(vector)

  def _ReplaceRegExGrouping(self, grouping: re.Match[str]) -> str:
    alternatives = grouping.group(1).split(",")
    return "(" + "|".join(re.escape(s) for s in alternatives) + ")"

  def _ReplaceRegExPart(self, part: str) -> str:
    if part == "**/":
      return "(?:.*\\/)?"
    elif part == "*":
      return "[^\\/]*"
    elif part == "?":
      return "[^\\/]"
    elif GROUPING_PATTERN.match(part):
      return GROUPING_PATTERN.sub(self._ReplaceRegExGrouping, part)
    else:
      return re.escape(part)

  def ExplainComponents(
      self,
      example_count: int,
      knowledge_base: knowledge_base_pb2.KnowledgeBase,
  ) -> Sequence[flows_pb2.GlobComponentExplanation]:
    """Returns a list of GlobComponentExplanations with examples."""
    parts = _COMPONENT_SPLIT_PATTERN.split(self._value)
    components = []

    for glob_part in parts:
      if not glob_part:
        continue

      component = flows_pb2.GlobComponentExplanation(glob_expression=glob_part)

      if GROUPING_PATTERN.match(glob_part):
        examples = self.InterpolateGrouping(glob_part)
      elif _VAR_PATTERN.match(glob_part):
        # Examples for variable substitutions might not be 100 % accurate,
        # because the scope is not shared between two variables. Thus,
        # if a GlobExpression uses %%users.a%% and %%users.b%%, the underlying
        # user might be different for a and b. For the sake of explaining
        # possible values, this should still be enough.
        interpolation = artifact_utils.KnowledgeBaseInterpolation(
            pattern=glob_part,
            kb=knowledge_base,
        )

        examples = interpolation.results
      else:
        examples = []

      component.examples.extend(list(itertools.islice(examples, example_count)))
      components.append(component)

    return components

  def AsRegEx(self) -> rdf_standard.RegularExpression:
    """Return the current glob as a simple regex.

    Note: No interpolation is performed.

    Returns:
      A RegularExpression() object.
    """
    parts = _REGEX_SPLIT_PATTERN.split(self._value)
    result = "".join(self._ReplaceRegExPart(p) for p in parts)

    return rdf_standard.RegularExpression("(?i)\\A%s\\Z" % result)



#!/usr/bin/env python
"""In this file we define constants that can be used by client and server."""

# Special folders we want to report back for each user. The format here is:
# registry key, folder name (relative to ProfileImagePath), protobuf name.
profile_folders = [
    ("AppData", "Application Data", "app_data"),
    (
        "Cache",
        "AppData\\Local\\Microsoft\\Windows\\Temporary Internet Files",
        "cache",
    ),
    ("Cookies", "Cookies", "cookies"),
    ("Desktop", "Desktop", "desktop"),
    ("Favorites", "Favorites", "favorites"),
    ("History", "AppData\\Local\\Microsoft\\Windows\\History", "history"),
    ("Local AppData", "AppData\\Roaming", "local_app_data"),
    ("My Music", "Music", "my_music"),
    ("My Pictures", "Pictures", "my_pictures"),
    ("My Video", "Videos", "my_video"),
    ("NetHood", "NetHood", "net_hood"),
    ("Personal", "Documents", "personal"),
    ("PrintHood", "PrintHood", "print_hood"),
    (
        "Programs",
        "AppData\\Roaming\\Microsoft\\Windows\\Start Menu\\Programs",
        "programs",
    ),
    ("Recent", "Recent", "recent"),
    ("SendTo", "SendTo", "send_to"),
    (
        "Start Menu",
        "AppData\\Roaming\\Microsoft\\Windows\\Start Menu",
        "start_menu",
    ),
    (
        "Startup",
        "AppData\\Roaming\\Microsoft\\Windows\\Start Menu\\Programs\\Startup",
        "startup",
    ),
    ("Templates", "Templates", "templates"),
]

CLIENT_MAX_BUFFER_SIZE = 640 * 1024



#!/usr/bin/env python
"""The various FileFinder rdfvalues."""

from grr_response_core.lib import rdfvalue
from grr_response_core.lib.rdfvalues import client as rdf_client
from grr_response_core.lib.rdfvalues import client_fs as rdf_client_fs
from grr_response_core.lib.rdfvalues import crypto as rdf_crypto
from grr_response_core.lib.rdfvalues import paths as rdf_paths
from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_proto import flows_pb2


class FileFinderModificationTimeCondition(rdf_structs.RDFProtoStruct):
  protobuf = flows_pb2.FileFinderModificationTimeCondition
  rdf_deps = [
      rdfvalue.RDFDatetime,
  ]


class FileFinderAccessTimeCondition(rdf_structs.RDFProtoStruct):
  protobuf = flows_pb2.FileFinderAccessTimeCondition
  rdf_deps = [
      rdfvalue.RDFDatetime,
  ]


class FileFinderInodeChangeTimeCondition(rdf_structs.RDFProtoStruct):
  protobuf = flows_pb2.FileFinderInodeChangeTimeCondition
  rdf_deps = [
      rdfvalue.RDFDatetime,
  ]


class FileFinderSizeCondition(rdf_structs.RDFProtoStruct):
  protobuf = flows_pb2.FileFinderSizeCondition


class FileFinderExtFlagsCondition(rdf_structs.RDFProtoStruct):

  protobuf = flows_pb2.FileFinderExtFlagsCondition

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)
    self.linux_bits_set = self.linux_bits_set or 0
    self.linux_bits_unset = self.linux_bits_unset or 0
    self.osx_bits_set = self.osx_bits_set or 0
    self.osx_bits_unset = self.osx_bits_unset or 0


class FileFinderContentsRegexMatchCondition(rdf_structs.RDFProtoStruct):
  protobuf = flows_pb2.FileFinderContentsRegexMatchCondition

  rdf_deps = [rdfvalue.RDFBytes]


class FileFinderContentsLiteralMatchCondition(rdf_structs.RDFProtoStruct):
  """An RDF value representing file finder contents literal match conditions."""

  protobuf = flows_pb2.FileFinderContentsLiteralMatchCondition

  rdf_deps = [rdfvalue.RDFBytes]

  def Validate(self):
    """Check the literal match condition is well constructed."""
    super().Validate()

    # The literal must not be empty in the literal match condition.
    if not self.HasField("literal") or not self.literal:
      raise ValueError(
          "No literal provided to FileFinderContentsLiteralMatchCondition."
      )


class FileFinderCondition(rdf_structs.RDFProtoStruct):
  """An RDF value representing file finder conditions."""

  protobuf = flows_pb2.FileFinderCondition
  rdf_deps = [
      FileFinderAccessTimeCondition,
      FileFinderContentsLiteralMatchCondition,
      FileFinderContentsRegexMatchCondition,
      FileFinderInodeChangeTimeCondition,
      FileFinderModificationTimeCondition,
      FileFinderSizeCondition,
      FileFinderExtFlagsCondition,
  ]

  @classmethod
  def AccessTime(cls, **kwargs):
    condition_type = cls.Type.ACCESS_TIME
    opts = FileFinderAccessTimeCondition(**kwargs)
    return cls(condition_type=condition_type, access_time=opts)

  @classmethod
  def ModificationTime(cls, **kwargs):
    condition_type = cls.Type.MODIFICATION_TIME
    opts = FileFinderModificationTimeCondition(**kwargs)
    return cls(condition_type=condition_type, modification_time=opts)

  @classmethod
  def InodeChangeTime(cls, **kwargs):
    condition_type = cls.Type.INODE_CHANGE_TIME
    opts = FileFinderInodeChangeTimeCondition(**kwargs)
    return cls(condition_type=condition_type, inode_change_time=opts)

  @classmethod
  def Size(cls, **kwargs):
    condition_type = cls.Type.SIZE
    opts = FileFinderSizeCondition(**kwargs)
    return cls(condition_type=condition_type, size=opts)

  @classmethod
  def ExtFlags(cls, **kwargs):
    condition_type = cls.Type.EXT_FLAGS
    opts = FileFinderExtFlagsCondition(**kwargs)
    return cls(condition_type=condition_type, ext_flags=opts)

  @classmethod
  def ContentsLiteralMatch(cls, **kwargs):
    condition_type = cls.Type.CONTENTS_LITERAL_MATCH
    opts = FileFinderContentsLiteralMatchCondition(**kwargs)
    return cls(condition_type=condition_type, contents_literal_match=opts)

  @classmethod
  def ContentsRegexMatch(cls, **kwargs):
    condition_type = cls.Type.CONTENTS_REGEX_MATCH
    opts = FileFinderContentsRegexMatchCondition(**kwargs)
    return cls(condition_type=condition_type, contents_regex_match=opts)

  def Validate(self):
    super().Validate()

    if self.HasField("modification_time"):
      self.modification_time.Validate()
    if self.HasField("access_time"):
      self.access_time.Validate()
    if self.HasField("inode_change_time"):
      self.inode_change_time.Validate()
    if self.HasField("size"):
      self.size.Validate()
    if self.HasField("ext_flags"):
      self.ext_flags.Validate()
    if self.HasField("contents_regex_match"):
      self.contents_regex_match.Validate()
    if self.HasField("contents_literal_match"):
      self.contents_literal_match.Validate()


class FileFinderStatActionOptions(rdf_structs.RDFProtoStruct):
  """FileFinder stat action options RDFStruct."""

  protobuf = flows_pb2.FileFinderStatActionOptions

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

    if not self.HasField("collect_ext_attrs"):
      self.collect_ext_attrs = False


class FileFinderHashActionOptions(rdf_structs.RDFProtoStruct):
  """FileFinder hash action options RDFStruct."""

  protobuf = flows_pb2.FileFinderHashActionOptions
  rdf_deps = [
      rdfvalue.ByteSize,
  ]

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

    if not self.HasField("collect_ext_attrs"):
      self.collect_ext_attrs = False


class FileFinderDownloadActionOptions(rdf_structs.RDFProtoStruct):
  """FileFinder download action options RDFStruct."""

  protobuf = flows_pb2.FileFinderDownloadActionOptions
  rdf_deps = [
      rdfvalue.ByteSize,
  ]

  def __init__(self, *args, **kwargs):
    super().__init__(*args, **kwargs)

    if not self.HasField("collect_ext_attrs"):
      self.collect_ext_attrs = False


class FileFinderAction(rdf_structs.RDFProtoStruct):
  """An RDF value describing a file-finder action."""

  protobuf = flows_pb2.FileFinderAction
  rdf_deps = [
      FileFinderDownloadActionOptions,
      FileFinderHashActionOptions,
      FileFinderStatActionOptions,
  ]

  @classmethod
  def Stat(cls, **kwargs):
    action_type = cls.Action.STAT
    opts = FileFinderStatActionOptions(**kwargs)
    return FileFinderAction(action_type=action_type, stat=opts)

  @classmethod
  def Hash(cls, **kwargs):
    action_type = cls.Action.HASH
    opts = FileFinderHashActionOptions(**kwargs)
    return FileFinderAction(action_type=action_type, hash=opts)

  @classmethod
  def Download(cls, **kwargs):
    action_type = cls.Action.DOWNLOAD
    opts = FileFinderDownloadActionOptions(**kwargs)
    return FileFinderAction(action_type=action_type, download=opts)


class FileFinderArgs(rdf_structs.RDFProtoStruct):
  """An RDF value representing file finder flow arguments."""

  protobuf = flows_pb2.FileFinderArgs
  rdf_deps = [
      FileFinderAction,
      FileFinderCondition,
      rdf_paths.GlobExpression,
  ]

  def Validate(self):
    super().Validate()

    for condition in self.conditions:
      condition.Validate()


class FileFinderResult(rdf_structs.RDFProtoStruct):
  protobuf = flows_pb2.FileFinderResult
  rdf_deps = [
      rdf_client.BufferReference,
      rdf_crypto.Hash,
      rdf_client_fs.StatEntry,
      rdf_client_fs.BlobImageDescriptor,
  ]


class CollectFilesByKnownPathArgs(rdf_structs.RDFProtoStruct):
  """Arguments for CollectFilesByKnownPath."""

  protobuf = flows_pb2.CollectFilesByKnownPathArgs
  rdf_deps = []


class CollectFilesByKnownPathResult(rdf_structs.RDFProtoStruct):
  """Result returned by CollectFilesByKnownPath."""

  protobuf = flows_pb2.CollectFilesByKnownPathResult
  rdf_deps = [
      rdf_crypto.Hash,
      rdf_client_fs.StatEntry,
  ]


class CollectFilesByKnownPathProgress(rdf_structs.RDFProtoStruct):
  """Progress returned by CollectFilesByKnownPath."""

  protobuf = flows_pb2.CollectFilesByKnownPathProgress
  rdf_deps = []


class CollectMultipleFilesArgs(rdf_structs.RDFProtoStruct):
  """Arguments for CollectMultipleFiles."""

  protobuf = flows_pb2.CollectMultipleFilesArgs
  rdf_deps = [
      rdf_paths.GlobExpression,
      FileFinderModificationTimeCondition,
      FileFinderAccessTimeCondition,
      FileFinderInodeChangeTimeCondition,
      FileFinderSizeCondition,
      FileFinderExtFlagsCondition,
      FileFinderContentsRegexMatchCondition,
      FileFinderContentsLiteralMatchCondition,
  ]


class CollectMultipleFilesResult(rdf_structs.RDFProtoStruct):
  """Result returned by CollectMultipleFiles."""

  protobuf = flows_pb2.CollectMultipleFilesResult
  rdf_deps = [
      rdf_crypto.Hash,
      rdf_client_fs.StatEntry,
  ]


class CollectMultipleFilesProgress(rdf_structs.RDFProtoStruct):
  """Progress returned by CollectMultipleFiles."""

  protobuf = flows_pb2.CollectMultipleFilesProgress
  rdf_deps = []


class StatMultipleFilesArgs(rdf_structs.RDFProtoStruct):
  """Arguments for StatMultipleFiles."""

  protobuf = flows_pb2.StatMultipleFilesArgs
  rdf_deps = [
      rdf_paths.GlobExpression,
      FileFinderModificationTimeCondition,
      FileFinderAccessTimeCondition,
      FileFinderInodeChangeTimeCondition,
      FileFinderSizeCondition,
      FileFinderExtFlagsCondition,
      FileFinderContentsRegexMatchCondition,
      FileFinderContentsLiteralMatchCondition,
  ]


class HashMultipleFilesArgs(rdf_structs.RDFProtoStruct):
  """Arguments for HashMultipleFiles."""

  protobuf = flows_pb2.HashMultipleFilesArgs
  rdf_deps = [
      rdf_paths.GlobExpression,
      FileFinderModificationTimeCondition,
      FileFinderAccessTimeCondition,
      FileFinderInodeChangeTimeCondition,
      FileFinderSizeCondition,
      FileFinderExtFlagsCondition,
      FileFinderContentsRegexMatchCondition,
      FileFinderContentsLiteralMatchCondition,
  ]


class HashMultipleFilesProgress(rdf_structs.RDFProtoStruct):
  """Progress returned by HashMultipleFiles."""

  protobuf = flows_pb2.HashMultipleFilesProgress
  rdf_deps = []



#!/usr/bin/env python
"""Configuration parameters for client builder and server packaging."""

import os
import time

from grr_response_core.lib import config_lib
from grr_response_core.lib import type_info

config_lib.DEFINE_string(
    name="ClientBuilder.output_extension",
    default=None,
    help="The file extension for the client (OS dependent).")

config_lib.DEFINE_string(
    name="ClientBuilder.package_dir",
    default="%(ClientBuilder.build_root_dir)/%(Client.name)-pkg",
    help="OSX package name.")

config_lib.DEFINE_bool(
    "ClientBuilder.fleetspeak_bundled", False,
    "Whether to bundle a Fleetspeak installation into the client package.")

config_lib.DEFINE_string(
    "ClientBuilder.fleetspeak_install_dir",
    "%(fleetspeak-client-bin@fleetspeak-client-bin|resource)",
    "Location of the bundled Fleetspeak installation.")

config_lib.DEFINE_string(
    "ClientBuilder.fleetspeak_client_config", "",
    "Location of the fleetspeak client configuration generated by "
    "fleetspeak-config.")

config_lib.DEFINE_string(
    "Client.fleetspeak_service_name", "FleetspeakService",
    "Name of the Fleetspeak (upstart, systemd or Windows) service. Used to "
    "restart the Fleetspeak service during GRR client installation.")

config_lib.DEFINE_string(
    "ClientBuilder.client_path",
    default="grr_response_client.client",
    help="Full module path for GRR client's main file.")

config_lib.DEFINE_string(
    "ClientBuilder.fleetspeak_service_dir", "/etc/fleetspeak/textservices",
    "Directory where Fleetspeak expects service configs to be. Only applies "
    "if Client.fleetspeak_enabled is true.")

config_lib.DEFINE_string(
    "ClientBuilder.fleetspeak_plist_path", None,
    "Path where the Fleetspeak client installs its plist file. Only applies "
    "if Client.fleetspeak_enabled is true.")

config_lib.DEFINE_string(
    "ClientBuilder.fleetspeak_config_path",
    "%(install_data/fleetspeak/%(Client.platform)/"
    "grr_service_config.txt.in@grr-response-core|resource)",
    "Path to GRR's Fleetspeak service configuration.")

config_lib.DEFINE_string(
    "ClientBuilder.version_ini_path", None,
    "Path to the version.ini file to be used when building a client template. "
    "If not specified, version.ini packaged with grr-response-core package "
    "will be used.")


class PathTypeInfo(type_info.String):
  """A path to a file or a directory."""

  def __init__(self, must_exist=True, **kwargs):
    self.must_exist = must_exist
    super().__init__(**kwargs)

  def Validate(self, value):
    value = super().Validate(value)
    if self.must_exist and not os.access(value, os.R_OK):
      raise type_info.TypeValueError(
          "Path %s does not exist for %s" % (value, self.name))

    return value

  def FromString(self, string):
    return os.path.normpath(string)


# PyInstaller build configuration.
config_lib.DEFINE_string(
    name="PyInstaller.spec",
    help="The spec file contents to use for building the client.",
    default=r"""
import glob
import os
import platform
import sys

# By default build in one dir mode.
client_path = r"%(%(%(ClientBuilder.client_path)|module_path)|fixpathsep)"

WINDOWS_IMPORTS = []
if platform.system\(\).lower\(\) == 'windows':
  WINDOWS_IMPORTS = ["win32process", "win32timezone"]

a = Analysis\(
    [client_path],
    # TODO\(https://github.com/pypa/setuptools/issues/1963\): py2_warn is
    # a workaround. Revisit in the future, whether this is needed.
    hiddenimports=WINDOWS_IMPORTS + ["pkg_resources.py2_warn"],
    hookspath=None\)

# Remove some optional libraries that would be packed but serve no purpose.
for prefix in ["IPython"]:
  for collection in [a.binaries, a.datas, a.pure]:
    for item in collection[:]:
      if item[0].startswith\(prefix\):
        collection.remove\(item\)

pyz = PYZ\(
    a.pure\)
exe = EXE\(
    pyz,
    a.scripts,
    exclude_binaries=1,
    name=os.path.join\("build", "grr-client"\),
    debug=False,
    strip=False,
    upx=False,
    embed_manifest=False,
    console=True,
    target_arch=r"%(PyInstaller.target_arch)" or None,
    version=os.path.join\(r"%(PyInstaller.build_dir)", "version.txt"\),
    icon=os.path.join\(r"%(PyInstaller.build_dir)", "grr.ico"\)\)

RESOURCES_PREFIX = os.path.join\(sys.prefix, "resources"\)

coll = COLLECT\(
    exe,
    a.binaries,
    a.zipfiles,
    a.datas,
    strip=False,
    upx=False,
    name="grr-client"
\)
""",
)

config_lib.DEFINE_string(
    name="PyInstaller.distpath",
    help=("Passed to PyInstaller as the --distpath flag. This sets the output "
          "directory for PyInstaller."),
    default="./dist")

config_lib.DEFINE_string(
    name="PyInstaller.version",
    help="The version.txt file contents to use for building the client.",
    default=r"""
VSVersionInfo\(
  ffi=FixedFileInfo\(
    filevers=\(%(Source.version_major), %(Source.version_minor),
               %(Source.version_revision), %(Source.version_release)\),
    prodvers=\(%(Source.version_major), %(Source.version_minor),
               %(Source.version_revision), %(Source.version_release)\),
    mask=0x3f,
    flags=0x0,
    OS=0x40004,
    fileType=0x1,
    subtype=0x0,
    date=\(0, 0\)
    \),
  kids=[
  VarFileInfo\([VarStruct\('Translation', [1033, 1200]\)]\)
  ]
\)
""")

config_lib.DEFINE_string("PyInstaller.icon_path",
                         "%(install_data/grr.ico@grr-response-core|resource)",
                         "A path to the icon to use for building the client.")

config_lib.DEFINE_string(
    "PyInstaller.build_dir",
    default="%(ClientBuilder.build_root_dir)/%(ClientBuilder.build_dest)",
    help="The path to the build directory.")

config_lib.DEFINE_string(
    "PyInstaller.dpkg_root",
    default="%(ClientBuilder.build_root_dir)/dist",
    help="Pyinstaller dpkg root.")

config_lib.DEFINE_string(
    "PyInstaller.workpath_dir",
    default="%(ClientBuilder.build_root_dir)/workpath",
    help="Pyinstaller working directory.")

config_lib.DEFINE_string(
    "PyInstaller.target_arch",
    default="",
    help=(
        "Pyinstaller target architecture (for macOS builds only). Defaults to "
        "a default PyInstaller behavior: producing non-unversal binary with "
        "the current machine's architecture. Set this option to 'universal2' "
        "to make PyInstaller produce universal binaries (you have to run on "
        "universal Python binary then). Set to 'x86_64' or 'arm64' to force "
        "a particular architecture. For details see: "
        "https://pyinstaller.org/en/stable/man/pyinstaller.html?highlight=universal2#mac-os-specific-options"
    ),
)

config_lib.DEFINE_string(
    name="Client.prefix",
    default="",
    help="A prefix for the client name, usually dbg_ for debug builds.")

config_lib.DEFINE_string(
    name="ClientBuilder.output_basename",
    default=("%(Client.prefix)%(Client.name)_"
             "%(Source.version_string)_%(Client.arch)"),
    help="The base name of the output package.")

# Windows client specific options.
config_lib.DEFINE_bool(
    "ClientBuilder.console",
    default=False,
    help="Should the application be built as a console program. "
    "This aids debugging in windows.")

config_lib.DEFINE_choice(
    name="ClientBuilder.build_type",
    default="Release",
    choices=["Release", "Debug"],
    help="Type of build (Debug, Release)")

config_lib.DEFINE_string(
    name="ClientBuilder.template_extension",
    default=".zip",
    help="The extension to appear on templates.")

config_lib.DEFINE_string(
    name="PyInstaller.template_basename",
    default=("%(Client.name)_%(Source.version_string)_%(Client.arch)"),
    help="The template name of the output package.")

config_lib.DEFINE_string(
    name="PyInstaller.template_filename",
    default=(
        "%(PyInstaller.template_basename)%(ClientBuilder.template_extension)"),
    help="The template file name of the output package.")

config_lib.DEFINE_option(
    PathTypeInfo(
        name="ClientBuilder.template_dir",
        must_exist=False,
        default=("%(grr-response-templates@grr-response-templates|resource)/"
                 "templates"),
        description="The directory holding executable template files."))

config_lib.DEFINE_option(
    PathTypeInfo(
        name="ClientBuilder.template_path",
        must_exist=False,
        default=(
            "%(ClientBuilder.template_dir)/%(PyInstaller.template_filename)"),
        description="The full path to the executable template files for "
        "building."))

config_lib.DEFINE_option(
    PathTypeInfo(
        name="ClientBuilder.executables_dir",
        must_exist=False,
        default="%(executables@grr-response-core|resource)",
        description="The path to the grr executables directory."))

config_lib.DEFINE_string(
    name="ClientBuilder.config_filename",
    default="%(Client.binary_name).yaml",
    help=("The name of the configuration file which will be embedded in the "
          "deployable binary."))

config_lib.DEFINE_string(
    name="ClientBuilder.maintainer",
    default="GRR <grr-dev@googlegroups.com>",
    help="The client package's maintainer.")

config_lib.DEFINE_string(
    name="ClientBuilder.debian_build_time",
    default=time.strftime("%a, %d %b %Y %H:%M:%S +0000", time.gmtime()),
    help="The build time put into the debian package. Needs to be formatted"
    " like the output of 'date -R'.")

config_lib.DEFINE_string(
    name="ClientBuilder.rpm_build_time",
    default=time.strftime("%a %b %d %Y", time.gmtime()),
    help="The build time put into the rpm package. Needs to be formatted"
    " according to the rpm specs.")

config_lib.DEFINE_string(
    name="ClientBuilder.debian_version",
    default="%(Template.version_numeric)-1",
    help="The version of the debian package.")

config_lib.DEFINE_string(
    name="ClientBuilder.debian_package_base",
    default=("%(ClientBuilder.package_name)_"
             "%(ClientBuilder.debian_version)_%(Template.arch)"),
    help="The filename of the debian package without extension.")

config_lib.DEFINE_string(
    name="ClientBuilder.package_name",
    default="%(Client.name)",
    help="The debian package name.")

config_lib.DEFINE_string(
    "ClientBuilder.vs_arch",
    default=None,
    help="Visual studio architecture string.")

config_lib.DEFINE_string(
    "ClientBuilder.vs_env_script",
    default=None,
    help="Path to visual studio environment variables bat file.")

config_lib.DEFINE_string(
    "ClientBuilder.vs_dir",
    default=r"%{C:\Program Files (x86)\Microsoft Visual Studio 14.0}",
    help="Path to visual studio installation dir.")

config_lib.DEFINE_string(
    "ClientBuilder.build_root_dir",
    default=None,
    help="Root directory for client builds.")

config_lib.DEFINE_string(
    "ClientBuilder.build_dest",
    default="%(Client.name)-build",
    help="Output directory for client building.")

config_lib.DEFINE_string(
    "ClientBuilder.install_dir",
    default="/usr/lib/%(Client.name)",
    help="Target installation directory for client builds.")

# Needed for package maker. Do not touch.
config_lib.DEFINE_string(
    "ClientBuilder.mangled_output_basename",
    default="%(Client.name)_%(Source.version_major)."
    "%(Source.version_minor).%(Source.version_revision)",
    help="OS X package maker mangled name.")

config_lib.DEFINE_string(
    "ClientBuilder.package_maker_organization",
    default=None,
    help="OS X package maker organization name.")

config_lib.DEFINE_string(
    "ClientBuilder.signing_cert_name",
    default=None,
    help="Name of a Darwin signing cert.")

config_lib.DEFINE_string(
    "ClientBuilder.signing_keychain_file",
    default="%(HOME|env)/Library/Keychains/MacApplicationSigning.keychain",
    help="Path to a keychain file to be used to sign Darwin binaries.")

config_lib.DEFINE_string(
    "ClientBuilder.target_dir",
    default=None,
    help="ClientBuilder target directory.")

config_lib.DEFINE_string(
    "ClientBuilder.daemon_link",
    default="/usr/sbin/%(Client.binary_name)",
    help="The installer package will create a link in the system to the "
    "installed binary.")

# These options will be used by client.client_build when running buildandrepack
# and can be used to customize what is built for each client label.
config_lib.DEFINE_multichoice(
    name="ClientBuilder.target_platforms",
    default=[],
    choices=[
        "darwin_amd64_dmg", "linux_amd64_deb", "linux_i386_deb",
        "linux_amd64_rpm", "linux_i386_rpm", "windows_amd64_exe",
        "windows_i386_exe"
    ],
    help="Platforms that will be built by client_build buildandrepack")

config_lib.DEFINE_string(
    "ClientBuilder.rpm_signing_key_public_keyfile",
    default="/etc/alternatives/grr_rpm_signing_key",
    help="Public key file for post-signing verification.")

config_lib.DEFINE_string(
    "ClientBuilder.rpm_gpg_name",
    default="GRR Team",
    help="The gpg name should match your gpg key name.")

# Parameters for signing Windows binaries on Linux hosts.
config_lib.DEFINE_string(
    "ClientBuilder.windows_signing_key",
    default="/etc/alternatives/grr_windows_signing_key",
    help="Path to GRR signing key. Should symlink "
    "to actual key.")

config_lib.DEFINE_string(
    "ClientBuilder.windows_signing_cert",
    default="/etc/alternatives/grr_windows_signing_cert",
    help="Path to GRR signing cert. Should symlink "
    "to actual cert.")

config_lib.DEFINE_string(
    "ClientBuilder.windows_signing_application_name",
    default="GRR",
    help="Signing cert application name.")

# Parameters for signing Windows binaries on Windows hosts.
config_lib.DEFINE_string(
    "ClientBuilder.signtool_signing_cmd",
    default=None,
    help="Full commandline for signtool to sign files.")

config_lib.DEFINE_string(
    "ClientBuilder.signtool_verification_cmd",
    default=None,
    help="Full commandline for signtool to verify signatures.")

config_lib.DEFINE_string(
    "ClientBuilder.extra_package_dependencies",
    default="",
    help="Extra dependencies, put verbatim into \"Depends:\" in debian/control "
    "or into \"Requires:\" in the RPM spec file.")

config_lib.DEFINE_string(
    "ClientBuilder.wix_tools_path",
    default="",
    help="Path to install location of wix toolset.")

config_lib.DEFINE_bool(
    "ClientBuilder.build_msi",
    default=False,
    help="If true, an MSI template is built on Windows.")

config_lib.DEFINE_string(
    name="ClientRepacker.output_basename",
    default=("%(Client.prefix)%(Client.name)_"
             "%(Template.version_string)_%(Template.arch)"),
    help="The base name of the output package.")

config_lib.DEFINE_option(
    PathTypeInfo(
        name="ClientRepacker.output_filename",
        must_exist=False,
        default=(
            "%(ClientRepacker.output_basename)%(ClientBuilder.output_extension)"
        ),
        description="The filename of the generated installer file."))

config_lib.DEFINE_option(
    PathTypeInfo(
        name="ClientRepacker.output_path",
        must_exist=False,
        default=("%(ClientBuilder.executables_dir)"
                 "/installers/%(ClientRepacker.output_filename)"),
        description="The full path to the generated installer file."))

# These values are determined from the template at repack time.
config_lib.DEFINE_choice(
    name="Template.build_type",
    default="Release",
    choices=["Release", "Debug"],
    help="Type of build (Debug, Release)")

config_lib.DEFINE_list(
    name="Template.build_context",
    default=[],
    help="List of build contexts that should be reapplied at repack.")

config_lib.DEFINE_integer("Template.version_major", None,
                          "Major version number of client template.")

config_lib.DEFINE_integer("Template.version_minor", None,
                          "Minor version number of client template.")

config_lib.DEFINE_integer("Template.version_revision", None,
                          "Revision number of client template.")

config_lib.DEFINE_integer("Template.version_release", None,
                          "Release number of client template.")

config_lib.DEFINE_string(
    "Template.version_string", "%(version_major).%(version_minor)."
    "%(version_revision).%(version_release)",
    "Version string of the client template.")

config_lib.DEFINE_integer(
    "Template.version_numeric", "%(version_major)%(version_minor)"
    "%(version_revision)%(version_release)",
    "Version string of the template as an integer.")

config_lib.DEFINE_string("Template.arch", None,
                         "The architecture of the client template.")



#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from grr_response_core.lib.rdfvalues import client as rdf_client
from grr_response_proto import jobs_pb2
from grr_response_proto import knowledge_base_pb2
from grr_response_proto import sysinfo_pb2


def ToProtoPackageRepository(
    rdf: rdf_client.PackageRepository,
) -> sysinfo_pb2.PackageRepository:
  return rdf.AsPrimitiveProto()


def ToRDFPackageRepository(
    proto: sysinfo_pb2.PackageRepository,
) -> rdf_client.PackageRepository:
  return rdf_client.PackageRepository.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoManagementAgent(
    rdf: rdf_client.ManagementAgent,
) -> sysinfo_pb2.ManagementAgent:
  return rdf.AsPrimitiveProto()


def ToRDFManagementAgent(
    proto: sysinfo_pb2.ManagementAgent,
) -> rdf_client.ManagementAgent:
  return rdf_client.ManagementAgent.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoPwEntry(rdf: rdf_client.PwEntry) -> knowledge_base_pb2.PwEntry:
  return rdf.AsPrimitiveProto()


def ToRDFPwEntry(proto: knowledge_base_pb2.PwEntry) -> rdf_client.PwEntry:
  return rdf_client.PwEntry.FromSerializedBytes(proto.SerializeToString())


def ToProtoGroup(rdf: rdf_client.Group) -> knowledge_base_pb2.Group:
  return rdf.AsPrimitiveProto()


def ToRDFGroup(proto: knowledge_base_pb2.Group) -> rdf_client.Group:
  return rdf_client.Group.FromSerializedBytes(proto.SerializeToString())


def ToProtoUserFromUser(rdf: rdf_client.User) -> knowledge_base_pb2.User:
  return rdf.AsPrimitiveProto()


def ToRDFUser(proto: knowledge_base_pb2.User) -> rdf_client.User:
  return rdf_client.User.FromSerializedBytes(proto.SerializeToString())


def ToProtoUserFromKnowledgeBaseUser(
    rdf: rdf_client.KnowledgeBaseUser,
) -> knowledge_base_pb2.User:
  return rdf.AsPrimitiveProto()


def ToRDFKnowledgeBaseUser(
    proto: knowledge_base_pb2.User,
) -> rdf_client.KnowledgeBaseUser:
  return rdf_client.KnowledgeBaseUser.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoKnowledgeBase(
    rdf: rdf_client.KnowledgeBase,
) -> knowledge_base_pb2.KnowledgeBase:
  return rdf.AsPrimitiveProto()


def ToRDFKnowledgeBase(
    proto: knowledge_base_pb2.KnowledgeBase,
) -> rdf_client.KnowledgeBase:
  return rdf_client.KnowledgeBase.FromSerializedBytes(proto.SerializeToString())


def ToProtoHardwareInfo(
    rdf: rdf_client.HardwareInfo,
) -> sysinfo_pb2.HardwareInfo:
  return rdf.AsPrimitiveProto()


def ToRDFHardwareInfo(
    proto: sysinfo_pb2.HardwareInfo,
) -> rdf_client.HardwareInfo:
  return rdf_client.HardwareInfo.FromSerializedBytes(proto.SerializeToString())


def ToProtoClientInformation(
    rdf: rdf_client.ClientInformation,
) -> jobs_pb2.ClientInformation:
  return rdf.AsPrimitiveProto()


def ToRDFClientInformation(
    proto: jobs_pb2.ClientInformation,
) -> rdf_client.ClientInformation:
  return rdf_client.ClientInformation.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoBufferReference(
    rdf: rdf_client.BufferReference,
) -> jobs_pb2.BufferReference:
  return rdf.AsPrimitiveProto()


def ToRDFBufferReference(
    proto: jobs_pb2.BufferReference,
) -> rdf_client.BufferReference:
  return rdf_client.BufferReference.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoProcess(rdf: rdf_client.Process) -> sysinfo_pb2.Process:
  return rdf.AsPrimitiveProto()


def ToRDFProcess(proto: sysinfo_pb2.Process) -> rdf_client.Process:
  return rdf_client.Process.FromSerializedBytes(proto.SerializeToString())


def ToProtoNamedPipe(rdf: rdf_client.NamedPipe) -> sysinfo_pb2.NamedPipe:
  return rdf.AsPrimitiveProto()


def ToRDFNamedPipe(proto: sysinfo_pb2.NamedPipe) -> rdf_client.NamedPipe:
  return rdf_client.NamedPipe.FromSerializedBytes(proto.SerializeToString())


def ToProtoSoftwarePackage(
    rdf: rdf_client.SoftwarePackage,
) -> sysinfo_pb2.SoftwarePackage:
  return rdf.AsPrimitiveProto()


def ToRDFSoftwarePackage(
    proto: sysinfo_pb2.SoftwarePackage,
) -> rdf_client.SoftwarePackage:
  return rdf_client.SoftwarePackage.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoSoftwarePackages(
    rdf: rdf_client.SoftwarePackages,
) -> sysinfo_pb2.SoftwarePackages:
  return rdf.AsPrimitiveProto()


def ToRDFSoftwarePackages(
    proto: sysinfo_pb2.SoftwarePackages,
) -> rdf_client.SoftwarePackages:
  return rdf_client.SoftwarePackages.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoLogMessage(rdf: rdf_client.LogMessage) -> jobs_pb2.LogMessage:
  return rdf.AsPrimitiveProto()


def ToRDFLogMessage(proto: jobs_pb2.LogMessage) -> rdf_client.LogMessage:
  return rdf_client.LogMessage.FromSerializedBytes(proto.SerializeToString())


def ToProtoUname(rdf: rdf_client.Uname) -> jobs_pb2.Uname:
  return rdf.AsPrimitiveProto()


def ToRDFUname(proto: jobs_pb2.Uname) -> rdf_client.Uname:
  return rdf_client.Uname.FromSerializedBytes(proto.SerializeToString())


def ToProtoStartupInfo(rdf: rdf_client.StartupInfo) -> jobs_pb2.StartupInfo:
  return rdf.AsPrimitiveProto()


def ToRDFStartupInfo(proto: jobs_pb2.StartupInfo) -> rdf_client.StartupInfo:
  return rdf_client.StartupInfo.FromSerializedBytes(proto.SerializeToString())


def ToProtoOSXServiceInformation(
    rdf: rdf_client.OSXServiceInformation,
) -> sysinfo_pb2.OSXServiceInformation:
  return rdf.AsPrimitiveProto()


def ToRDFOSXServiceInformation(
    proto: sysinfo_pb2.OSXServiceInformation,
) -> rdf_client.OSXServiceInformation:
  return rdf_client.OSXServiceInformation.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoRunKey(rdf: rdf_client.RunKey) -> sysinfo_pb2.RunKey:
  return rdf.AsPrimitiveProto()


def ToRDFRunKey(proto: sysinfo_pb2.RunKey) -> rdf_client.RunKey:
  return rdf_client.RunKey.FromSerializedBytes(proto.SerializeToString())


def ToProtoClientCrash(rdf: rdf_client.ClientCrash) -> jobs_pb2.ClientCrash:
  return rdf.AsPrimitiveProto()


def ToRDFClientCrash(proto: jobs_pb2.ClientCrash) -> rdf_client.ClientCrash:
  return rdf_client.ClientCrash.FromSerializedBytes(proto.SerializeToString())


def ToProtoEdrAgent(rdf: rdf_client.EdrAgent) -> jobs_pb2.EdrAgent:
  return rdf.AsPrimitiveProto()


def ToRDFEdrAgent(proto: jobs_pb2.EdrAgent) -> rdf_client.EdrAgent:
  return rdf_client.EdrAgent.FromSerializedBytes(proto.SerializeToString())


def ToProtoFleetspeakValidationInfoTag(
    rdf: rdf_client.FleetspeakValidationInfoTag,
) -> jobs_pb2.FleetspeakValidationInfoTag:
  return rdf.AsPrimitiveProto()


def ToRDFFleetspeakValidationInfoTag(
    proto: jobs_pb2.FleetspeakValidationInfoTag,
) -> rdf_client.FleetspeakValidationInfoTag:
  return rdf_client.FleetspeakValidationInfoTag.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoFleetspeakValidationInfo(
    rdf: rdf_client.FleetspeakValidationInfo,
) -> jobs_pb2.FleetspeakValidationInfo:
  return rdf.AsPrimitiveProto()


def ToRDFFleetspeakValidationInfo(
    proto: jobs_pb2.FleetspeakValidationInfo,
) -> rdf_client.FleetspeakValidationInfo:
  return rdf_client.FleetspeakValidationInfo.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoClientSummary(
    rdf: rdf_client.ClientSummary,
) -> jobs_pb2.ClientSummary:
  return rdf.AsPrimitiveProto()


def ToRDFClientSummary(
    proto: jobs_pb2.ClientSummary,
) -> rdf_client.ClientSummary:
  return rdf_client.ClientSummary.FromSerializedBytes(proto.SerializeToString())



#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from grr_response_core.lib.rdfvalues import timeline as rdf_timeline
from grr_response_proto import timeline_pb2


def ToProtoTimelineArgs(
    rdf: rdf_timeline.TimelineArgs,
) -> timeline_pb2.TimelineArgs:
  return rdf.AsPrimitiveProto()


def ToRDFTimelineArgs(
    proto: timeline_pb2.TimelineArgs,
) -> rdf_timeline.TimelineArgs:
  return rdf_timeline.TimelineArgs.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoTimelineResult(
    rdf: rdf_timeline.TimelineResult,
) -> timeline_pb2.TimelineResult:
  return rdf.AsPrimitiveProto()


def ToRDFTimelineResult(
    proto: timeline_pb2.TimelineResult,
) -> rdf_timeline.TimelineResult:
  return rdf_timeline.TimelineResult.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoTimelineEntry(
    rdf: rdf_timeline.TimelineEntry,
) -> timeline_pb2.TimelineEntry:
  return rdf.AsPrimitiveProto()


def ToRDFTimelineEntry(
    proto: timeline_pb2.TimelineEntry,
) -> rdf_timeline.TimelineEntry:
  return rdf_timeline.TimelineEntry.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoTimelineProgress(
    rdf: rdf_timeline.TimelineProgress,
) -> timeline_pb2.TimelineProgress:
  return rdf.AsPrimitiveProto()


def ToRDFTimelineProgress(
    proto: timeline_pb2.TimelineProgress,
) -> rdf_timeline.TimelineProgress:
  return rdf_timeline.TimelineProgress.FromSerializedBytes(
      proto.SerializeToString()
  )



#!/usr/bin/env python
"""Provides conversion functions to be used during RDFProtoStruct migration."""

from grr_response_core.lib.rdfvalues import crypto as rdf_crypto
from grr_response_proto import jobs_pb2


def ToProtoCertificate(rdf: rdf_crypto.Certificate) -> jobs_pb2.Certificate:
  return rdf.AsPrimitiveProto()


def ToRDFCertificate(proto: jobs_pb2.Certificate) -> rdf_crypto.Certificate:
  return rdf_crypto.Certificate.FromSerializedBytes(proto.SerializeToString())


def ToProtoHash(rdf: rdf_crypto.Hash) -> jobs_pb2.Hash:
  return rdf.AsPrimitiveProto()


def ToRDFHash(proto: jobs_pb2.Hash) -> rdf_crypto.Hash:
  return rdf_crypto.Hash.FromSerializedBytes(proto.SerializeToString())


def ToProtoSignedBlob(rdf: rdf_crypto.SignedBlob) -> jobs_pb2.SignedBlob:
  return rdf.AsPrimitiveProto()


def ToRDFSignedBlob(proto: jobs_pb2.SignedBlob) -> rdf_crypto.SignedBlob:
  return rdf_crypto.SignedBlob.FromSerializedBytes(proto.SerializeToString())


def ToProtoSymmetricCipher(
    rdf: rdf_crypto.SymmetricCipher,
) -> jobs_pb2.SymmetricCipher:
  return rdf.AsPrimitiveProto()


def ToRDFSymmetricCipher(
    proto: jobs_pb2.SymmetricCipher,
) -> rdf_crypto.SymmetricCipher:
  return rdf_crypto.SymmetricCipher.FromSerializedBytes(
      proto.SerializeToString()
  )


def ToProtoPassword(rdf: rdf_crypto.Password) -> jobs_pb2.Password:
  return rdf.AsPrimitiveProto()


def ToRDFPassword(proto: jobs_pb2.Password) -> rdf_crypto.Password:
  return rdf_crypto.Password.FromSerializedBytes(proto.SerializeToString())



#!/usr/bin/env python
"""A module with RDF wrappers for large file collection proto messages."""

from grr_response_core.lib.rdfvalues import paths as rdf_paths
from grr_response_core.lib.rdfvalues import structs as rdf_structs
from grr_response_proto import large_file_pb2


class CollectLargeFileArgs(rdf_structs.RDFProtoStruct):
  """An RDF wrapper for arguments of the large file collection action."""

  protobuf = large_file_pb2.CollectLargeFileArgs
  rdf_deps = [
      rdf_paths.PathSpec,
  ]


class CollectLargeFileResult(rdf_structs.RDFProtoStruct):
  """An RDF wrapper for result of the large file collection action."""

  protobuf = large_file_pb2.CollectLargeFileResult
  rdf_deps = []



#!/usr/bin/env python
"""Settings for ACLs/approvals system."""


from grr_response_core.lib import config_lib

config_lib.DEFINE_string(
    "ACL.approvers_config_file", "%(Config.directory)/approvers.yaml",
    "File that defines who can approve access to "
    "clients with certain labels.")

config_lib.DEFINE_integer("ACL.approvers_required", 2,
                          "The number of approvers required for access.")

config_lib.DEFINE_string(
    "ACL.group_access_manager_class", "NoGroupAccess",
    "This class handles interfacing with corporate group"
    "directories for granting access. Override with a "
    "class that understands your LDAP/AD/whatever setup.")

config_lib.DEFINE_integer(
    "ACL.token_expiry",
    28 * 24 * 60 * 60,
    "The default duration in seconds of a valid approval token. "
    "Default of four weeks.",
)

config_lib.DEFINE_integer(
    "ACL.token_max_expiry",
    367 * 24 * 60 * 60,
    "The maximum duration in seconds of a valid approval token. "
    "Default 1 leap year and 1 day.",
)



#!/usr/bin/env python
"""Tests for (de)serialization logic."""

from absl.testing import absltest

from grr_response_core.lib import serialization


class BoolConverterTest(absltest.TestCase):

  def testFromHumanReadableTrue(self):
    self.assertIs(serialization.FromHumanReadable(bool, "true"), True)
    self.assertIs(serialization.FromHumanReadable(bool, "True"), True)
    self.assertIs(serialization.FromHumanReadable(bool, "TRUE"), True)
    self.assertIs(serialization.FromHumanReadable(bool, "1"), True)

  def testFromHumanReadableFalse(self):
    self.assertIs(serialization.FromHumanReadable(bool, "false"), False)
    self.assertIs(serialization.FromHumanReadable(bool, "False"), False)
    self.assertIs(serialization.FromHumanReadable(bool, "FALSE"), False)
    self.assertIs(serialization.FromHumanReadable(bool, "0"), False)

  def testFromHumanReadableRaisesOnIncorrectInteger(self):
    with self.assertRaises(ValueError):
      serialization.FromHumanReadable(bool, "2")

  def testFromHumanReadableRaisesOnWeirdInput(self):
    with self.assertRaises(ValueError):
      serialization.FromHumanReadable(bool, "yes")

  def testWireFormat(self):
    self.assertIs(
        serialization.FromWireFormat(bool, serialization.ToWireFormat(True)),
        True,
    )
    self.assertIs(
        serialization.FromWireFormat(bool, serialization.ToWireFormat(False)),
        False,
    )

  def testBytes(self):
    self.assertIs(
        serialization.FromBytes(bool, serialization.ToBytes(True)), True
    )
    self.assertIs(
        serialization.FromBytes(bool, serialization.ToBytes(False)), False
    )

  def testHumanReadable(self):
    self.assertIs(serialization.FromHumanReadable(bool, str(True)), True)
    self.assertIs(serialization.FromHumanReadable(bool, str(False)), False)


if __name__ == "__main__":
  absltest.main()



#!/usr/bin/env python
"""API config options."""


from grr_response_core.lib import config_lib
from grr_response_core.lib import rdfvalue

config_lib.DEFINE_integer("API.DailyFlowRequestLimit", "10",
                          "Number of flows a user can run on a single client "
                          "per day before being blocked by throttling. Set to "
                          "0 to disable checking.")

config_lib.DEFINE_semantic_value(
    rdfvalue.Duration,
    "API.FlowDuplicateInterval",
    default="1200s",
    help="Amount of time "
    "that needs to pass before the throttler will allow "
    "an identical flow to run on the same client. Set "
    "to 0s to disable checking.")

config_lib.DEFINE_string("API.RouterACLConfigFile", "",
                         "The file containing API acls, see "
                         "grr/config/api_acls.yaml for an example.")

config_lib.DEFINE_string("API.DefaultRouter", "DisabledApiCallRouter",
                         "The default router used by the API if there are no "
                         "rules defined in API.RouterACLConfigFile or if none "
                         "of these rules matches.")



#!/usr/bin/env python
"""This file contains various utility classes used by GRR."""

from collections.abc import Iterable
import errno
import functools
import getpass
import gzip
import io
import os
import pathlib
import queue
import random
import re
import shlex
import shutil
import socket
import struct
import tarfile
import tempfile
import threading
import time
from typing import Generic, Optional, TypeVar
import weakref
import zipfile

import psutil

from grr_response_core.lib.util import precondition


class Error(Exception):
  pass


class ParsingError(Error):
  pass


def Proxy(f):
  """A helper to create a proxy method in a class."""

  def Wrapped(self, *args):
    return getattr(self, f)(*args)

  return Wrapped


class TempDirectory:
  """A self cleaning temporary directory.

  Do not use this function for any client related temporary files! Use
  the functionality provided by client_actions/tempfiles.py instead.
  """

  def __enter__(self):
    self.name = tempfile.mkdtemp()

    return self.name

  def __exit__(self, exc_type, exc_value, traceback):
    shutil.rmtree(self.name, True)


# This is a synchronize decorator.
def Synchronized(f):
  """Synchronization decorator."""

  @functools.wraps(f)
  def NewFunction(self, *args, **kw):
    with self.lock:
      return f(self, *args, **kw)

  return NewFunction


class InterruptableThread(threading.Thread):
  """A class which exits once the main thread exits."""

  def __init__(
      self,
      target=None,
      args=None,
      kwargs=None,
      sleep_time=10,
      name: Optional[str] = None,
      **kw,
  ):
    self.exit = False
    self.last_run = 0
    self.target = target
    self.args = args or ()
    self.kwargs = kwargs or {}
    self.sleep_time = sleep_time
    if name is None:
      raise ValueError("Please name your threads.")

    # TODO(hanuszczak): Incorrect type specification for the `name` param.
    # pytype: disable=wrong-arg-count
    super().__init__(name=name, **kw)
    # pytype: enable=wrong-arg-count

    # Do not hold up program exit
    self.daemon = True

  def Iterate(self):
    """This will be repeatedly called between sleeps."""

  def Stop(self):
    self.exit = True

  def run(self):
    # When the main thread exits, the time module might disappear and be already
    # None. We take a local reference to the functions we need.
    sleep = time.sleep
    now = time.time

    while not self.exit:
      if self.target:
        self.target(*self.args, **self.kwargs)
      else:
        self.Iterate()

      # Implement interruptible sleep here.
      self.last_run = now()

      # Exit if the main thread disappears.
      while time and not self.exit and now() < self.last_run + self.sleep_time:
        sleep(1)


class Node:
  """An entry to a linked list."""

  next = None
  prev = None
  data = None

  def __init__(self, key, data):
    self.data = data
    self.key = key

  def __str__(self):
    return "Node %s: %s" % (self.key, self.data)

  def __repr__(self):
    return str(self)


# TODO(user):pytype: self.next and self.prev are assigned to self but then
# are used in AppendNode in a very different way. Should be redesigned.
# pytype: disable=attribute-error
class LinkedList:
  """A simple doubly linked list used for fast caches."""

  def __init__(self):
    # We are the head node.
    self.next = self.prev = self
    self.size = 0

  def AppendNode(self, node):
    self.size += 1
    last_node = self.prev

    last_node.next = node
    node.prev = last_node
    node.next = self
    self.prev = node

  def PopLeft(self):
    """Returns the head node and removes it from the list."""
    if self.next is self:
      raise IndexError("Pop from empty list.")

    first_node = self.next
    self.Unlink(first_node)
    return first_node

  def Pop(self):
    """Returns the tail node and removes it from the list."""
    if self.prev is self:
      raise IndexError("Pop from empty list.")

    last_node = self.prev
    self.Unlink(last_node)
    return last_node

  def Unlink(self, node):
    """Removes a given node from the list."""
    self.size -= 1

    node.prev.next = node.next
    node.next.prev = node.prev
    node.next = node.prev = None

  def __iter__(self):
    p = self.next
    while p is not self:
      yield p
      p = p.next

  def __len__(self):
    return self.size

  def __str__(self):
    return "[" + ", ".join(map(str, self)) + "]"

  def Print(self):
    p = self.next
    while p is not self:
      print("%s: prev %r next %r\n" % (p.data, p.prev, p.next))
      p = p.next


# pytype: enable=attribute-error


class FastStore:
  """This is a cache which expires objects in oldest first manner.

  This implementation first appeared in PyFlag.
  """

  def __init__(self, max_size=10):
    """Constructor.

    Args:
       max_size: The maximum number of objects held in cache.
    """
    # This class implements a LRU cache which needs fast updates of the LRU
    # order for random elements. This is usually implemented by using a
    # dict for fast lookups and a linked list for quick deletions / insertions.
    self._age = LinkedList()
    self._hash = {}
    self._limit = max_size
    self.lock = threading.RLock()

  def KillObject(self, obj):
    """Perform cleanup on objects when they expire.

    Should be overridden by classes which need to perform special cleanup.
    Args:
      obj: The object which was stored in the cache and is now expired.
    """

  @Synchronized
  def __iter__(self):
    return iter([(key, n.data) for key, n in self._hash.items()])

  @Synchronized
  def Expire(self):
    """Expires old cache entries."""
    while len(self._age) > self._limit:
      node = self._age.PopLeft()
      self._hash.pop(node.key, None)
      self.KillObject(node.data)

  @Synchronized
  def Put(self, key, obj):
    """Add the object to the cache."""
    # Remove the old entry if it is there.
    node = self._hash.pop(key, None)
    if node:
      self._age.Unlink(node)

    # Make a new node and insert it.
    node = Node(key=key, data=obj)
    self._hash[key] = node
    self._age.AppendNode(node)

    self.Expire()

    return key

  @Synchronized
  def ExpireObject(self, key):
    """Expire a specific object from cache."""
    node = self._hash.pop(key, None)
    if node:
      self._age.Unlink(node)
      self.KillObject(node.data)

      return node.data

  @Synchronized
  def ExpireRegEx(self, regex):
    """Expire all the objects with the key matching the regex."""
    reg = re.compile(regex)
    for key in list(self._hash):
      if reg.match(key):
        self.ExpireObject(key)

  @Synchronized
  def ExpirePrefix(self, prefix):
    """Expire all the objects with the key having a given prefix."""
    for key in list(self._hash):
      if key.startswith(prefix):
        self.ExpireObject(key)

  @Synchronized
  def Pop(self, key):
    """Remove the object from the cache completely."""
    node = self._hash.get(key)

    if node:
      del self._hash[key]
      self._age.Unlink(node)

      return node.data

  @Synchronized
  def Get(self, key):
    """Fetch the object from cache.

    Objects may be flushed from cache at any time. Callers must always
    handle the possibility of KeyError raised here.

    Args:
      key: The key used to access the object.

    Returns:
      Cached object.

    Raises:
      KeyError: If the object is not present in the cache.
    """
    if key not in self._hash:
      raise KeyError(key)

    node = self._hash[key]

    self._age.Unlink(node)
    self._age.AppendNode(node)

    return node.data

  @Synchronized
  def __contains__(self, obj):
    return obj in self._hash

  @Synchronized
  def __getitem__(self, key):
    return self.Get(key)

  @Synchronized
  def Flush(self):
    """Flush all items from cache."""
    while self._age:
      node = self._age.PopLeft()
      self.KillObject(node.data)

    self._hash = dict()

  def __len__(self):
    return len(self._hash)


_T = TypeVar("_T")


class TimeBasedCacheEntry(Generic[_T]):

  def __init__(self, timestamp: float, value: _T):
    self.timestamp = timestamp
    self.value: _T = value


class TimeBasedCache(FastStore):
  """A Cache which expires based on time."""

  active_caches = None
  house_keeper_thread = None

  def __init__(self, max_size=10, max_age=600):
    """Constructor.

    This cache will refresh the age of the cached object as long as they are
    accessed within the allowed age. The age refers to the time since it was
    last touched.

    Args:
      max_size: The maximum number of objects held in cache.
      max_age: The maximum length of time an object is considered alive.
    """
    super().__init__(max_size)
    self.max_age = max_age

    def HouseKeeper():
      """A housekeeper thread which expunges old objects."""
      if not time:
        # This might happen when the main thread exits, we don't want to raise.
        return

      now = time.time()

      for cache in TimeBasedCache.active_caches:
        # Only expunge while holding the lock on the data store.
        with cache.lock:
          # pylint: disable=protected-access
          # We need to take a copy of the value list because we are changing
          # this dict during the iteration.
          for node in list(cache._hash.values()):
            # Expire the object if it is too old.
            if node.data.timestamp + cache.max_age < now:
              cache.KillObject(node.data)

              cache._age.Unlink(node)
              cache._hash.pop(node.key, None)
          # pylint: enable=protected-access

    if not TimeBasedCache.house_keeper_thread:
      TimeBasedCache.active_caches = weakref.WeakSet()
      # This thread is designed to never finish.
      TimeBasedCache.house_keeper_thread = InterruptableThread(
          name="HouseKeeperThread", target=HouseKeeper
      )
      TimeBasedCache.house_keeper_thread.start()
    TimeBasedCache.active_caches.add(self)

  @Synchronized
  def Get(self, key):
    now = time.time()
    stored = super().Get(key)
    if stored.timestamp + self.max_age < now:
      raise KeyError("Expired")

    # This updates the timestamp in place to keep the object alive
    stored.timestamp = now

    return stored.value

  def Put(self, key, obj):
    super().Put(key, TimeBasedCacheEntry(time.time(), obj))


class AgeBasedCache(TimeBasedCache):
  """A cache which holds objects for a maximum length of time.

  This differs from the TimeBasedCache which keeps the objects alive as long as
  they are accessed.
  """

  @Synchronized
  def Get(self, key):
    now = time.time()
    stored = FastStore.Get(self, key)
    if stored.timestamp + self.max_age < now:
      raise KeyError("Expired")

    return stored.value


class Struct:
  """A baseclass for parsing binary Structs."""

  # Derived classes must initialize this into an array of (format,
  # name) tuples.
  _fields: list[tuple[str, str]]

  def __init__(self, data):
    """Parses ourselves from data."""
    format_str = "".join([x[0] for x in self._fields])
    self.size = struct.calcsize(format_str)

    try:
      parsed_data = struct.unpack(format_str, data[: self.size])

    except struct.error:
      raise ParsingError("Unable to parse")

    for i in range(len(self._fields)):
      setattr(self, self._fields[i][1], parsed_data[i])

  def __repr__(self):
    """Produce useful text representation of the Struct."""
    dat = []
    for _, name in self._fields:
      dat.append("%s=%s" % (name, getattr(self, name)))
    return "%s(%s)" % (self.__class__.__name__, ", ".join(dat))

  @classmethod
  def GetSize(cls):
    """Calculate the size of the struct."""
    format_str = "".join([x[0] for x in cls._fields])
    return struct.calcsize(format_str)


def SmartUnicode(string):
  """Returns a unicode object.

  This function will always return a unicode object. It should be used to
  guarantee that something is always a unicode object.

  Args:
    string: The string to convert.

  Returns:
    a unicode object.
  """
  if isinstance(string, str):
    return string

  if isinstance(string, bytes):
    return string.decode("utf-8", "ignore")

  return str(string)


def Xor(bytestr, key):
  """Returns a `bytes` object where each byte has been xored with key."""
  precondition.AssertType(bytestr, bytes)
  return bytes([byte ^ key for byte in bytestr])


def FormatAsHexString(num, width=None, prefix="0x"):
  """Takes an int and returns the number formatted as a hex string."""
  # Strip "0x".
  hex_str = hex(num)[2:]
  # Strip "L" for long values.
  hex_str = hex_str.replace("L", "")
  if width:
    hex_str = hex_str.rjust(width, "0")
  return "%s%s" % (prefix, hex_str)


def FormatAsTimestamp(timestamp: int) -> str:
  if not timestamp:
    return "-"

  return time.strftime("%Y-%m-%d %H:%M:%S", time.gmtime(timestamp))


def NormalizePath(path: str, sep: str = "/") -> str:
  """A sane implementation of os.path.normpath.

  The standard implementation treats leading / and // as different leading to
  incorrect normal forms.

  NOTE: Its ok to use a relative path here (without leading /) but any /../ will
  still be removed anchoring the path at the top level (e.g. foo/../../../../bar
  => bar).

  Args:
     path: The path to normalize.
     sep: Separator used.

  Returns:
     A normalized path. In this context normalized means that all input paths
     that would result in the system opening the same physical file will produce
     the same normalized path.
  """
  precondition.AssertType(path, str)
  precondition.AssertType(sep, str)

  if not path:
    return sep

  path_list = path.split(sep)

  # This is a relative path and the first element is . or ..
  if path_list[0] in [".", "..", ""]:
    path_list.pop(0)

  # Deliberately begin at index 1 to preserve a single leading /
  i = 0

  while True:
    list_len = len(path_list)

    # We begin at the last known good position so we never iterate over path
    # elements which are already examined
    for i in range(i, len(path_list)):
      # Remove /./ form
      if path_list[i] == "." or not path_list[i]:
        path_list.pop(i)
        break

      # Remove /../ form
      elif path_list[i] == "..":
        path_list.pop(i)
        # Anchor at the top level
        if (i == 1 and path_list[0]) or i > 1:
          i -= 1
          path_list.pop(i)
        break

    # If we didn't alter the path so far we can quit
    if len(path_list) == list_len:
      return sep + sep.join(path_list)


# TODO(hanuszczak): The linter complains for a reason here, the signature of
# this function should be fixed as soon as possible.
def JoinPath(stem: str = "", *parts: str) -> str:  # pylint: disable=keyword-arg-before-vararg
  """A sane version of os.path.join.

  The intention here is to append the stem to the path. The standard module
  removes the path if the stem begins with a /.

  Args:
     stem: The stem to join to.
     *parts: parts of the path to join. The first arg is always the root and
       directory traversal is not allowed.

  Returns:
     a normalized path.
  """
  precondition.AssertIterableType(parts, str)
  precondition.AssertType(stem, str)

  result = (stem + NormalizePath("/".join(parts))).replace("//", "/")
  result = result.rstrip("/")

  return result or "/"


def ShellQuote(value):
  """Escapes the string for the safe use inside shell command line."""
  return shlex.quote(SmartUnicode(value))


def Join(*parts):
  """Join (AFF4) paths without normalizing.

  A quick join method that can be used to express the precondition that
  the parts are already normalized.

  Args:
    *parts: The parts to join

  Returns:
    The joined path.
  """

  return "/".join(parts)


def GeneratePassphrase(length=20):
  """Create a 20 char passphrase with easily typeable chars."""
  valid_chars = "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
  valid_chars += "0123456789 ,-_&$#"
  return "".join(random.choice(valid_chars) for i in range(length))


def PassphraseCallback(
    verify=False, prompt1="Enter passphrase:", prompt2="Verify passphrase:"
):
  """A utility function to read a passphrase from stdin."""
  while 1:
    try:
      p1 = getpass.getpass(prompt1)
      if verify:
        p2 = getpass.getpass(prompt2)
        if p1 == p2:
          break
      else:
        break
    except KeyboardInterrupt:
      return None
  return p1.encode("utf-8")


def FormatNumberAsString(num):
  """Return a large number in human readable form."""
  for suffix in ["b", "KB", "MB", "GB"]:
    if num < 1024.0:
      return "%3.2f%s" % (num, suffix)
    num /= 1024.0
  return "%3.1f%s" % (num, "TB")


class NotAValue:
  pass


class HeartbeatQueue(queue.Queue):
  """A queue that periodically calls a provided callback while waiting."""

  def __init__(self, *args, callback=None, **kw):
    queue.Queue.__init__(self, *args, **kw)
    self.callback = callback or (lambda: None)

  def get(self, poll_interval=5):
    while True:
      try:
        message = queue.Queue.get(self, block=True, timeout=poll_interval)
        break

      except queue.Empty:
        self.callback()

    return message


class RollingMemoryStream(object):
  """Append-only memory stream that allows writing data in chunks."""

  def __init__(self):
    self._stream = io.BytesIO()
    self._offset = 0

  def write(self, b):  # pylint: disable=invalid-name
    if not self._stream:
      raise ArchiveAlreadyClosedError("Attempting to write to a closed stream.")

    self._stream.write(b)
    self._offset += len(b)

  def flush(self):  # pylint: disable=invalid-name
    pass

  def tell(self):  # pylint: disable=invalid-name
    return self._offset

  def close(self):  # pylint: disable=invalid-name
    self._stream = None

  def GetValueAndReset(self):
    """Gets stream buffer since the last GetValueAndReset() call."""
    if not self._stream:
      raise ArchiveAlreadyClosedError(
          "Attempting to get a value from a closed stream."
      )

    value = self._stream.getvalue()
    self._stream.seek(0)
    self._stream.truncate()

    return value


class ArchiveAlreadyClosedError(Error):
  pass


# TODO(hanuszczak): Typings for `ZipFile` are ill-typed.
# pytype: disable=attribute-error,wrong-arg-types
class StreamingZipGenerator(object):
  """A streaming zip generator that can archive file-like objects."""

  def __init__(self, compression=zipfile.ZIP_STORED):
    self._compression = compression
    self._stream = RollingMemoryStream()
    self._zipfile = zipfile.ZipFile(
        self._stream, mode="w", compression=compression, allowZip64=True
    )

    self._zipopen = None

  def __enter__(self):
    return self

  def __exit__(self, exc_type, exc_value, traceback):
    del exc_type, exc_value, traceback  # Unused.
    return self.Close()

  def __del__(self):
    if self._zipopen is not None:
      self._zipopen.__exit__(None, None, None)
      self._zipopen = None

  # TODO(hanuszczak): Because we have to use `ZipFile::open`, there is no way to
  # specify per-file compression and write custom stat entry (but it should be
  # relevant only for dates). Once we remove this class and switch back to the
  # native implementation, it should be possible to fill this information again.
  def WriteFileHeader(self, arcname=None, compress_type=None, st=None):
    del compress_type, st  # Unused.

    if not self._stream:
      raise ArchiveAlreadyClosedError()

    self._zipopen = self._zipfile.open(arcname, mode="w")
    self._zipopen.__enter__()

    return self._stream.GetValueAndReset()

  def WriteFileChunk(self, chunk):
    precondition.AssertType(chunk, bytes)

    if not self._stream:
      raise ArchiveAlreadyClosedError()

    self._zipopen.write(chunk)

    return self._stream.GetValueAndReset()

  def WriteFileFooter(self):
    if not self._stream:
      raise ArchiveAlreadyClosedError()

    self._zipopen.__exit__(None, None, None)
    self._zipopen = None

    return self._stream.GetValueAndReset()

  def Close(self):
    if self._zipopen is not None:
      self._zipopen.__exit__(None, None, None)
      self._zipopen = None

    self._zipfile.close()

    value = self._stream.GetValueAndReset()
    self._stream.close()
    return value

  def WriteFromFD(self, src_fd, arcname=None, compress_type=None, st=None):
    """A convenience method for adding an entire file to the ZIP archive."""
    yield self.WriteFileHeader(
        arcname=arcname, compress_type=compress_type, st=st
    )

    while True:
      buf = src_fd.read(1024 * 1024)
      if not buf:
        break

      yield self.WriteFileChunk(buf)

    yield self.WriteFileFooter()

  @property
  def is_file_write_in_progress(self) -> bool:
    return bool(self._zipopen)

  @property
  def output_size(self):
    return self._stream.tell()


# pytype: enable=attribute-error,wrong-arg-types


class StreamingTarGenerator(object):
  """A streaming tar generator that can archive file-like objects."""

  FILE_CHUNK_SIZE = 1024 * 1024 * 4

  def __init__(self):
    super().__init__()

    self._stream = RollingMemoryStream()
    self._stream_gzip = gzip.GzipFile(mode="wb", fileobj=self._stream)

    self._ResetState()

  def _ResetState(self):
    self.cur_file_size = 0
    self.cur_info = None

  def __enter__(self):
    return self

  def __exit__(self, unused_type, unused_value, unused_traceback):
    self.Close()

  def Close(self):
    self._stream_gzip.close()

    value = self._stream.GetValueAndReset()
    self._stream.close()

    return value

  def WriteFileHeader(self, arcname=None, st=None):
    """Writes file header."""
    precondition.AssertType(arcname, str)

    if st is None:
      raise ValueError("Stat object can't be None.")

    self.cur_file_size = 0

    self.cur_info = tarfile.TarInfo()
    self.cur_info.type = tarfile.REGTYPE
    self.cur_info.name = arcname
    self.cur_info.size = st.st_size
    self.cur_info.mode = st.st_mode
    self.cur_info.mtime = st.st_mtime or time.time()

    self._stream_gzip.write(self.cur_info.tobuf(encoding="utf-8"))

    return self._stream.GetValueAndReset()

  def WriteFileChunk(self, chunk):
    """Writes file chunk."""

    self._stream_gzip.write(chunk)
    self.cur_file_size += len(chunk)
    return self._stream.GetValueAndReset()

  def WriteFileFooter(self):
    """Writes file footer (finishes the file)."""
    if self.cur_info is None:
      raise ValueError("WriteFileHeader() must be called first.")

    if self.cur_file_size != self.cur_info.size:
      raise IOError(
          "Incorrect file size: st_size=%d, but written %d bytes."
          % (self.cur_info.size, self.cur_file_size)
      )

    remainder = self.cur_file_size % tarfile.BLOCKSIZE
    if remainder > 0:
      self._stream_gzip.write(tarfile.NUL * (tarfile.BLOCKSIZE - remainder))

    self._ResetState()

    return self._stream.GetValueAndReset()

  @property
  def is_file_write_in_progress(self):
    return self.cur_info

  def WriteFromFD(self, src_fd, arcname=None, st=None):
    """Write an archive member from a file like object.

    Args:
      src_fd: A file like object, must support seek(), tell(), read().
      arcname: The name in the archive this should take.
      st: A stat object to be used for setting headers.

    Raises:
      ValueError: If st is omitted.
      ArchiveAlreadyClosedError: If the archive was already closed.
      IOError: if file size reported in st is different from the one that
          was actually read from the src_fd.

    Yields:
      Chunks of binary data.
    """

    yield self.WriteFileHeader(arcname=arcname, st=st)

    while 1:
      buf = src_fd.read(1024 * 1024)
      if not buf:
        break
      yield self.WriteFileChunk(buf)

    yield self.WriteFileFooter()

  @property
  def output_size(self):
    return self._stream.tell()


class Stubber:
  """A context manager for doing simple stubs."""

  def __init__(self, module, target_name, stub):
    self.target_name = target_name
    self.module = module
    self.stub = stub

  def __enter__(self):
    self.Start()

  def Stop(self):
    setattr(self.module, self.target_name, self.old_target)

  def Start(self):
    self.old_target = getattr(self.module, self.target_name, None)
    try:
      self.stub.old_target = self.old_target
    except AttributeError:
      pass
    setattr(self.module, self.target_name, self.stub)

  def __exit__(self, unused_type, unused_value, unused_traceback):
    self.Stop()


class MultiStubber(object):
  """A context manager for doing simple stubs."""

  def __init__(self, *args):
    self.stubbers = [Stubber(*x) for x in args]

  def Start(self):
    for x in self.stubbers:
      x.Start()

  def Stop(self):
    for x in self.stubbers:
      x.Stop()

  def __enter__(self):
    self.Start()

  def __exit__(self, t, value, traceback):
    self.Stop()


def EnsureDirExists(path):
  """Equivalent of makedir -p."""
  try:
    os.makedirs(path)
  except OSError as exc:
    # Necessary so we don't hide other errors such as permission denied.
    if exc.errno == errno.EEXIST and os.path.isdir(path):
      pass
    else:
      raise


def MergeDirectories(src: str, dst: str) -> None:
  """Merges the src directory tree into the dst directory tree."""
  src_dir = pathlib.Path(src)
  dst_dir = pathlib.Path(dst)
  for path in src_dir.glob("**/*"):
    if path.is_dir():
      continue
    relative_path = path.relative_to(src_dir)
    dst_path = dst_dir / relative_path
    EnsureDirExists(str(dst_path.parent))
    shutil.copy(str(path), str(dst_path))


def ResolveHostnameToIP(host, port):
  """Resolves a hostname to an IP address."""
  ip_addrs = socket.getaddrinfo(
      host, port, socket.AF_UNSPEC, 0, socket.IPPROTO_TCP
  )
  # getaddrinfo returns tuples (family, socktype, proto, canonname, sockaddr).
  # We are interested in sockaddr which is in turn a tuple
  # (address, port) for IPv4 or (address, port, flow info, scope id)
  # for IPv6. In both cases, we want the first element, the address.
  return ip_addrs[0][4][0]


# TODO: This module is way too big right now. It should be split
# into several smaller ones (such as `util.paths`, `util.collections` etc.).


def ProcessIdString():
  return "%s@%s:%d" % (
      psutil.Process().name(),
      socket.gethostname(),
      os.getpid(),
  )


def RegexListDisjunction(regex_list: Iterable[bytes]):
  precondition.AssertIterableType(regex_list, bytes)
  return b"(" + b")|(".join(regex_list) + b")"


def ReadFileBytesAsUnicode(file_obj):
  data = file_obj.read()
  precondition.AssertType(data, bytes)

  return data.decode("utf-8")


def RunOnce(fn):
  """Returns a decorated function that will only pass through the first call.

  At first execution, the return value or raised Exception is passed through and
  cached. Further calls will not be passed to `fn` and will return or raise
  the result of the first call.

  Be cautious when returning an Iterator, Generator or mutable value, since the
  result is shared by reference among all calls.

  Args:
    fn: The function to be decorated.

  Returns:
    A decorated function that will pass through only the first call.
  """

  @functools.wraps(fn)
  def _OneTimeFunction(*args, **kwargs):
    """Wrapper function that only passes through the first call."""
    if not _OneTimeFunction.executed:
      _OneTimeFunction.executed = True
      try:
        _OneTimeFunction.result = fn(*args, **kwargs)
      except BaseException as e:  # pylint: disable=broad-except
        _OneTimeFunction.exception = e
        raise  # Preserve original stack trace during first invocation.

    if _OneTimeFunction.exception is None:
      return _OneTimeFunction.result
    else:
      raise _OneTimeFunction.exception

  _OneTimeFunction.executed = False
  _OneTimeFunction.exception = None
  _OneTimeFunction.result = None
  return _OneTimeFunction
